@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 # system imports
 import ast
 import collections
@@ -17,7 +19,7 @@ import typing
 import sys
 
 # third part imports
-import numpy
+import numpy.typing
 
 # local libraries
 from nion.swift.model import PlugInManager
@@ -34,11 +36,19 @@ from nion.utils import Geometry
 
 if typing.TYPE_CHECKING:
     from nion.swift import DocumentController
+    from nion.ui import DrawingContext
+    from nion.ui import UserInterface
+
+_NDArray = numpy.typing.NDArray[typing.Any]
 
 _ = gettext.gettext
 
 
-def pose_get_string_message_box(ui, message_column, caption, text, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None):
+def pose_get_string_message_box(ui: UserInterface.UserInterface, message_column: UserInterface.BoxWidget, caption: str,
+                                text: str, accepted_fn: typing.Callable[[str], None],
+                                rejected_fn: typing.Optional[typing.Callable[[], None]] = None,
+                                accepted_text: typing.Optional[str] = None,
+                                rejected_text: typing.Optional[str] = None) -> UserInterface.BoxWidget:
     if accepted_text is None:
         accepted_text = _("OK")
     if rejected_text is None:
@@ -50,13 +60,13 @@ def pose_get_string_message_box(ui, message_column, caption, text, accepted_fn,
     caption_row.add_stretch()
     inside_row = ui.create_row_widget()
 
-    def reject_button_clicked():
-        if rejected_fn:
+    def reject_button_clicked() -> typing.Any:
+        if callable(rejected_fn):
             rejected_fn()
         return False
 
-    def accept_button_clicked():
-        accepted_fn(string_edit_widget.text)
+    def accept_button_clicked() -> typing.Any:
+        accepted_fn(string_edit_widget.text or str())
         return False
 
     string_edit_widget = ui.create_line_edit_widget()
@@ -85,19 +95,24 @@ def pose_get_string_message_box(ui, message_column, caption, text, accepted_fn,
     return message_box_widget
 
 
-def pose_confirmation_message_box(ui, message_column, caption, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None, display_rejected=True):
+def pose_confirmation_message_box(ui: UserInterface.UserInterface, message_column: UserInterface.BoxWidget,
+                                  caption: str, accepted_fn: typing.Callable[[], None],
+                                  rejected_fn: typing.Optional[typing.Callable[[], None]] = None,
+                                  accepted_text: typing.Optional[str] = None,
+                                  rejected_text: typing.Optional[str] = None,
+                                  display_rejected: bool = True) -> UserInterface.BoxWidget:
     if accepted_text is None:
         accepted_text = _("OK")
     if rejected_text is None:
         rejected_text = _("Cancel")
     message_box_widget = ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}
 
-    def reject_button_clicked():
+    def reject_button_clicked() -> typing.Any:
         if rejected_fn:
             rejected_fn()
         return False
 
-    def accept_button_clicked():
+    def accept_button_clicked() -> typing.Any:
         accepted_fn()
         return False
 
@@ -122,7 +137,7 @@ def pose_confirmation_message_box(ui, message_column, caption, accepted_fn, reje
 
 
 class ScriptListItem:
-    def __init__(self, full_path: str, indent: int = 0, indent_level: int = 0, show_dirname: bool = True):
+    def __init__(self, full_path: str, indent: int = 0, indent_level: int = 0, show_dirname: bool = True) -> None:
         self.__full_path = os.path.abspath(full_path)
         self._exists = False
         self.indent = indent
@@ -151,11 +166,11 @@ class ScriptListItem:
         return self._exists
 
     # Used by "sort"
-    def __lt__(self, other) -> bool:
+    def __lt__(self, other: typing.Any) -> bool:
         if isinstance(other, FolderListItem):
             return False
         if isinstance(other, ScriptListItem):
-            return locale.strxfrm(self.basename) < locale.strxfrm(other.basename)
+            return locale.strxfrm(self.basename.casefold()) < locale.strxfrm(other.basename.casefold())
         return NotImplemented
 
 
@@ -171,7 +186,7 @@ class FolderListItem(ScriptListItem):
     def content(self) -> typing.List[ScriptListItem]:
         return self.__content
 
-    def update_content_from_file_system(self, filter_pattern: typing.Optional[str] = None):
+    def update_content_from_file_system(self, filter_pattern: typing.Optional[str] = None) -> None:
         if os.path.isdir(self.full_path):
             dirlist = os.listdir(self.full_path)
             filtered_items = list()
@@ -183,15 +198,15 @@ class FolderListItem(ScriptListItem):
             self.__content = filtered_items
 
     # Used by "sort"
-    def __lt__(self, other) -> bool:
+    def __lt__(self, other: typing.Any) -> bool:
         if isinstance(other, FolderListItem):
-            return locale.strxfrm(self.basename) < locale.strxfrm(other.basename)
+            return locale.strxfrm(self.basename.casefold()) < locale.strxfrm(other.basename.casefold())
         if isinstance(other, ScriptListItem):
             return True
         return NotImplemented
 
 
-def _build_sorted_scripts_list(scripts_list: typing.List[ScriptListItem]) -> typing.List[ScriptListItem]:
+def _build_sorted_scripts_list(scripts_list: typing.Sequence[typing.Any]) -> typing.Sequence[ScriptListItem]:
     filtered_items = []
     for item in scripts_list:
         if item.indent_level == 0:
@@ -202,7 +217,7 @@ def _build_sorted_scripts_list(scripts_list: typing.List[ScriptListItem]) -> typ
         set_items.append(item)
         if isinstance(item, FolderListItem):
             if not item.folder_closed:
-                for content_item in item.content:
+                for content_item in sorted(item.content):
                     set_items.append(content_item)
     return set_items
 
@@ -218,7 +233,8 @@ def open_location(location: str) -> None:
 
 class ScriptListCanvasItemDelegate(Widgets.ListCanvasItemDelegate):
 
-    def __init__(self, ui, document_controller, update_list_fn: typing.Callable[[], typing.Any]):
+    def __init__(self, ui: UserInterface.UserInterface, document_controller: DocumentController.DocumentController,
+                 update_list_fn: typing.Callable[[], typing.Any]) -> None:
         super().__init__()
         self.__ui = ui
         self.__document_controller = document_controller
@@ -260,7 +276,7 @@ class ScriptListCanvasItemDelegate(Widgets.ListCanvasItemDelegate):
     def _minor_font_color() -> str:
         return "#888"
 
-    def mouse_pressed_in_item(self, mouse_index: int, pos: Geometry.IntPoint, modifiers) -> bool:
+    def mouse_pressed_in_item(self, mouse_index: int, pos: Geometry.IntPoint, modifiers: UserInterface.KeyboardModifiers) -> bool:
         display_item = self.items[mouse_index]
         width = self.__ui.get_font_metrics(self._major_font_size(), self._closed_folder_icon_strings()[0]).width
         if isinstance(display_item, FolderListItem) and display_item.indent - 4 < pos.x < display_item.indent + width + 2:
@@ -271,23 +287,25 @@ class ScriptListCanvasItemDelegate(Widgets.ListCanvasItemDelegate):
             return True
         return False
 
-    def context_menu_event(self, index: int, x: int, y: int, gx: int, gy: int) -> bool:
-        display_item = self.items[index]
-        menu = self.__document_controller.create_context_menu()
-        if isinstance(display_item, FolderListItem):
-            menu.add_menu_item(_("Open Folder"), functools.partial(open_location, display_item.full_path))
-        elif isinstance(display_item, ScriptListItem):
+    def context_menu_event(self, index: typing.Optional[int], x: int, y: int, gx: int, gy: int) -> bool:
+        if index is not None:
+            display_item = self.items[index]
             menu = self.__document_controller.create_context_menu()
-            menu.add_menu_item(_("Open Containing Folder"), functools.partial(open_location, display_item.dirname))
-        menu.popup(gx, gy)
-        return True
+            if isinstance(display_item, FolderListItem):
+                menu.add_menu_item(_("Open Folder"), functools.partial(open_location, display_item.full_path))
+            elif isinstance(display_item, ScriptListItem):
+                menu = self.__document_controller.create_context_menu()
+                menu.add_menu_item(_("Open Containing Folder"), functools.partial(open_location, display_item.dirname))
+            menu.popup(gx, gy)
+            return True
+        return False
 
-    def __calculate_indent(self, display_item):
+    def __calculate_indent(self, display_item: typing.Any) -> int:
         # An item that can cause an indent_level > 0 is always an open folder
         triangle_string, icon_string = self._open_folder_icon_strings()
-        return 4 + self.__ui.get_font_metrics(self._major_font_size(), triangle_string + icon_string).width * display_item.indent_level
+        return round(4 + self.__ui.get_font_metrics(self._major_font_size(), triangle_string + icon_string).width * display_item.indent_level)
 
-    def paint_item(self, drawing_context, display_item, rect, is_selected):
+    def paint_item(self, drawing_context: DrawingContext.DrawingContext, display_item: typing.Any, rect: Geometry.IntRect, is_selected: bool) -> None:
         if isinstance(display_item, FolderListItem):
             if display_item.folder_closed:
                 triangle_string, icon_string = self._closed_folder_icon_strings()
@@ -339,11 +357,9 @@ class RunScriptDialog(Dialog.ActionDialog):
 
         self.script_filter_pattern = "\\.py$"
 
-        self._create_menus()
-
         self.__cancelled = False
 
-        self.__thread = None
+        self.__thread: typing.Optional[threading.Thread] = None
 
         properties = dict()
         properties["min-height"] = 180
@@ -387,15 +403,16 @@ class RunScriptDialog(Dialog.ActionDialog):
             assert self.__profile
             add_dir = self.ui.get_persistent_string("import_directory", "")
             file_paths, filter_str, directory = self.get_file_paths_dialog(_("Add Scripts"), add_dir, "Python Files (*.py)", "Python Files (*.py)")
-            self.ui.set_persistent_string("import_directory", directory)
-            items = self.scripts_list_widget.items
-            for file_path_str in file_paths:
-                script_item = Profile.FileScriptItem(pathlib.Path(file_path_str))
-                self.__profile.append_script_item(script_item)
-                script_list_item = ScriptListItem(file_path_str)
-                script_list_item.script_item = script_item
-                items.append(script_list_item)
-            self.update_scripts_list(items)
+            if len(file_paths) > 0:
+                self.ui.set_persistent_string("import_directory", directory)
+                items = list(self.scripts_list_widget.items)
+                for file_path_str in file_paths:
+                    script_item = Profile.FileScriptItem(pathlib.Path(file_path_str))
+                    self.__profile.append_script_item(script_item)
+                    script_list_item = ScriptListItem(file_path_str)
+                    script_list_item.script_item = script_item
+                    items.append(script_list_item)
+                self.update_scripts_list(items)
 
         def add_folder_clicked() -> None:
             assert self.__profile
@@ -408,7 +425,7 @@ class RunScriptDialog(Dialog.ActionDialog):
                 if full_path not in sys.path:
                     sys.path.append(full_path)
                     self.__new_path_entries.append(full_path)
-                items = self.scripts_list_widget.items
+                items = list(self.scripts_list_widget.items)
                 script_item = Profile.FolderScriptItem(pathlib.Path(existing_directory))
                 self.__profile.append_script_item(script_item)
                 folder_list_item.script_item = script_item
@@ -512,7 +529,7 @@ class RunScriptDialog(Dialog.ActionDialog):
 
         self.__is_closed = False
 
-    def close(self):
+    def close(self) -> None:
         self.__is_closed = True
         for sync_event in self.__sync_events:
             sync_event.set()
@@ -525,10 +542,10 @@ class RunScriptDialog(Dialog.ActionDialog):
         super().close()
 
     @property
-    def cancelled(self):
+    def cancelled(self) -> bool:
         return self.__cancelled
 
-    def update_scripts_list(self, new_scripts_list):
+    def update_scripts_list(self, new_scripts_list: typing.Sequence[typing.Any]) -> None:
         for item in new_scripts_list:
             if isinstance(item, FolderListItem):
                 item.update_content_from_file_system(filter_pattern=self.script_filter_pattern)
@@ -537,11 +554,11 @@ class RunScriptDialog(Dialog.ActionDialog):
             item.check_existence()
         self.scripts_list_widget.items = items
 
-    def rebuild_scripts_list(self):
+    def rebuild_scripts_list(self) -> None:
         self.update_scripts_list(self.scripts_list_widget.items)
 
-    def __make_cancel_row(self):
-        def cancel_script():
+    def __make_cancel_row(self) -> UserInterface.BoxWidget:
+        def cancel_script() -> None:
             self.__cancelled = True
 
         cancel_button = self.ui.create_push_button_widget(_("Cancel"))
@@ -558,7 +575,7 @@ class RunScriptDialog(Dialog.ActionDialog):
         for module in current_modules:
             for path in self.__new_path_entries:
                 try:
-                    do_reload = os.path.abspath(module.__file__)[:len(path)] == path
+                    do_reload = os.path.abspath(module.__file__ or str())[:len(path)] == path
                 except (AttributeError, ValueError, TypeError):
                     pass
                 else:
@@ -571,7 +588,7 @@ class RunScriptDialog(Dialog.ActionDialog):
                         self.continue_after_parse_error(script_path)
                         return
 
-        self.__output_widget.text = None
+        self.__output_widget.text = str()
 
         script_name = os.path.basename(script_path)
 
@@ -589,11 +606,11 @@ class RunScriptDialog(Dialog.ActionDialog):
             return
 
         class AddCallFunctionNodeTransformer(ast.NodeTransformer):
-            def __init__(self, func_id, arg_id):
+            def __init__(self, func_id: str, arg_id: str) -> None:
                 self.__func_id = func_id
                 self.__arg_id = arg_id
 
-            def visit_Module(self, node):
+            def visit_Module(self, node: typing.Any) -> typing.Any:
                 name_expr = ast.Name(id=self.__func_id, ctx=ast.Load())
                 arg_expr = ast.Name(id=self.__arg_id, ctx=ast.Load())
                 call_expr = ast.Expr(value=ast.Call(func=name_expr, args=[arg_expr], keywords=[]))
@@ -609,28 +626,28 @@ class RunScriptDialog(Dialog.ActionDialog):
 
         compiled = compile(script_ast, script_name, 'exec')
 
-        def run_it(compiled, interactive_session):
+        def run_it(compiled: typing.Any, interactive_session: typing.Any) -> None:
             class APIBroker:
-                def get_interactive(self, version=None):
+                def get_interactive(self, version: typing.Optional[str] = None) -> typing.Any:
                     actual_version = "1.0.0"
-                    if Utility.compare_versions(version, actual_version) > 0:
+                    if Utility.compare_versions(version or str(), actual_version) > 0:
                         raise NotImplementedError("API requested version %s is greater than %s." % (version, actual_version))
                     return interactive_session
 
-                def get_api(self, version, ui_version=None):
+                def get_api(self, version: str, ui_version: typing.Optional[str] = None) -> typing.Any:
                     ui_version = ui_version if ui_version else "~1.0"
                     api = PlugInManager.api_broker_fn(version, ui_version)
-                    queued_api = FacadeQueued.API(api, None)
-                    queued_api._queue_task = api.queue_task
+                    queued_api = FacadeQueued.API(api, None)  # type: ignore
+                    queued_api._queue_task = api.queue_task  # type: ignore
                     return queued_api
 
-                def get_ui(self, version):
+                def get_ui(self, version: str) -> Declarative.DeclarativeUI:
                     actual_version = "1.0.0"
                     if Utility.compare_versions(version, actual_version) > 0:
                         raise NotImplementedError("API requested version %s is greater than %s." % (version, actual_version))
                     return Declarative.DeclarativeUI()
             try:
-                g = dict()
+                g: typing.Dict[str, typing.Any] = dict()
                 g["api_broker"] = APIBroker()
                 g["print"] = self.print
                 g["input"] = self.get_string
@@ -639,17 +656,17 @@ class RunScriptDialog(Dialog.ActionDialog):
                 self.__cancelled = False  # Reset cancelled flag to make "Run again" work after a script was cancelled
 
                 class StdoutCatcher:
-                    def __init__(self):
+                    def __init__(self) -> None:
                         pass
 
-                    def write(self, stuff):
+                    def write(self, stuff: str) -> None:
                         print_fn(stuff.rstrip())
 
-                    def flush(self):
+                    def flush(self) -> None:
                         pass
 
                 stdout = StdoutCatcher()
-                with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stdout):
+                with contextlib.redirect_stdout(typing.cast(typing.Any, stdout)), contextlib.redirect_stderr(typing.cast(typing.Any, stdout)):
                     exec(compiled, g)
             except Exception:
                 self.print("{}: {}".format(_("Error"), traceback.format_exc()))
@@ -658,8 +675,8 @@ class RunScriptDialog(Dialog.ActionDialog):
 
         self.run(script_path, functools.partial(run_it, compiled))
 
-    def continue_after_parse_error(self, script_path):
-        def func_continue():
+    def continue_after_parse_error(self, script_path: str) -> None:
+        def func_continue() -> None:
             result = self.confirm(_("Finished"), _("Run Again"), _("Close"))
 
             with self.__lock:
@@ -672,8 +689,8 @@ class RunScriptDialog(Dialog.ActionDialog):
         self.__thread = threading.Thread(target=func_continue)
         self.__thread.start()
 
-    def run(self, script_path, func):
-        def func_run(func):
+    def run(self, script_path: str, func: typing.Callable[[RunScriptDialog], None]) -> None:
+        def func_run(func: typing.Callable[[RunScriptDialog], None]) -> None:
             try:
                 func(self)
             except Exception:
@@ -692,7 +709,7 @@ class RunScriptDialog(Dialog.ActionDialog):
         self.__thread = threading.Thread(target=func_run, args=(func,))
         self.__thread.start()
 
-    def __handle_output_and_q(self):
+    def __handle_output_and_q(self) -> None:
         func = None
         with self.__lock:
             while len(self.__output_queue) > 0:
@@ -703,32 +720,32 @@ class RunScriptDialog(Dialog.ActionDialog):
         if callable(func):
             func()
 
-    def print(self, text):
+    def print(self, text: typing.Any) -> None:
         with self.__lock:
             self.__output_queue.append(str(text))
             self.document_controller.add_task("print_" + str(id(self)), self.__handle_output_and_q)
 
-    def print_debug(self, text):
+    def print_debug(self, text: typing.Any) -> None:
         self.print(text)
 
-    def print_info(self, text):
+    def print_info(self, text: typing.Any) -> None:
         self.print(text)
 
-    def print_warn(self, text):
+    def print_warn(self, text: typing.Any) -> None:
         self.print(text)
 
-    def get_string(self, prompt, default_str=None) -> typing.Optional[str]:
+    def get_string(self, prompt: str, default_str: typing.Optional[str] = None) -> typing.Optional[str]:
         """Return a string value that the user enters. Raises exception for cancel."""
         with self.sync_event() as accept_event:
             result = None
 
-            def perform():
-                def accepted(text):
+            def perform() -> None:
+                def accepted(text: str) -> None:
                     nonlocal result
                     result = text
                     accept_event.set()
 
-                def rejected():
+                def rejected() -> None:
                     accept_event.set()
 
                 self.__message_column.remove_all()
@@ -741,7 +758,7 @@ class RunScriptDialog(Dialog.ActionDialog):
             if self.__is_closed:
                 raise Exception("Cancel")
 
-        def update_message_column():
+        def update_message_column() -> None:
             self.__message_column.remove_all()
             self.__message_column.add(self.__make_cancel_row())
         self.document_controller.add_task("ui_" + str(id(self)), update_message_column)
@@ -752,17 +769,17 @@ class RunScriptDialog(Dialog.ActionDialog):
     def get_integer(self, prompt: str, default_value: int = 0) -> int:
         converter = Converter.IntegerToStringConverter()
         result = self.get_string(prompt, converter.convert(default_value))
-        return converter.convert_back(result)
+        return converter.convert_back(result) or 0
 
-    def get_float(self, prompt: str, default_value: float = 0, format_str: str = None) -> float:
+    def get_float(self, prompt: str, default_value: float = 0, format_str: typing.Optional[str] = None) -> float:
         converter = Converter.FloatToStringConverter(format_str)
         result = self.get_string(prompt, converter.convert(default_value))
-        return converter.convert_back(result)
+        return converter.convert_back(result) or 0.0
 
-    def show_ndarray(self, data: numpy.ndarray, title: str = None) -> None:
+    def show_ndarray(self, data: _NDArray, title: typing.Optional[str] = None) -> None:
         with self.sync_event() as accept_event:
 
-            def perform():
+            def perform() -> None:
                 result_display_panel = self.document_controller.next_result_display_panel()
                 if result_display_panel:
                     data_item = self.document_controller.add_data(data, title)
@@ -785,24 +802,24 @@ class RunScriptDialog(Dialog.ActionDialog):
         self.__sync_events.remove(sync_event)
 
     @contextlib.contextmanager
-    def sync_event(self):
+    def sync_event(self) -> typing.Iterator[threading.Event]:
         sync_event = threading.Event()
         self.__register_sync_event(sync_event)
         yield sync_event
         self.__unregister_sync_event(sync_event)
 
-    def __accept_reject(self, prompt, accepted_text, rejected_text, display_rejected):
+    def __accept_reject(self, prompt: str, accepted_text: typing.Optional[str], rejected_text: typing.Optional[str], display_rejected: bool) -> bool:
         """Return a boolean value for accept/reject."""
         with self.sync_event() as accept_event:
             result = False
 
-            def perform():
-                def accepted():
+            def perform() -> None:
+                def accepted() -> None:
                     nonlocal result
                     result = True
                     accept_event.set()
 
-                def rejected():
+                def rejected() -> None:
                     nonlocal result
                     result = False
                     accept_event.set()
@@ -818,7 +835,7 @@ class RunScriptDialog(Dialog.ActionDialog):
             if self.__is_closed:
                 raise Exception("Cancel")
 
-        def update_message_column():
+        def update_message_column() -> None:
             self.__message_column.remove_all()
             self.__message_column.add(self.__make_cancel_row())
         self.document_controller.add_task("ui_" + str(id(self)), update_message_column)
@@ -833,5 +850,5 @@ class RunScriptDialog(Dialog.ActionDialog):
     def confirm(self, prompt: str, accepted_text: str, rejected_text: str) -> bool:
         return self.__accept_reject(prompt, accepted_text, rejected_text, True)
 
-    def alert(self, prompt: str, button_label: str = None) -> None:
+    def alert(self, prompt: str, button_label: typing.Optional[str] = None) -> None:
         self.__accept_reject(prompt, button_label, None, False)
