@@ -6,6 +6,8 @@ import functools
 import gettext
 import random
 import string
+import threading
+import time
 import typing
 import uuid
 import weakref
@@ -15,8 +17,11 @@ import weakref
 
 # local libraries
 from nion.swift import DisplayPanel
+from nion.swift import FilterPanel
 from nion.swift import MimeTypes
+from nion.swift import Panel
 from nion.swift import Undo
+from nion.swift.model import Persistence
 from nion.swift.model import Utility
 from nion.swift.model import WorkspaceLayout
 from nion.ui import CanvasItem
@@ -24,8 +29,6 @@ from nion.ui import UserInterface
 
 if typing.TYPE_CHECKING:
     from nion.swift import DocumentController
-    from nion.swift import FilterPanel
-    from nion.swift import Panel
     from nion.swift.model import DisplayItem
     from nion.swift.model import DocumentModel
     from nion.swift.model import Project
@@ -34,22 +37,22 @@ if typing.TYPE_CHECKING:
 _ = gettext.gettext
 
 
-def create_image_desc():
+def create_image_desc() -> Persistence.PersistentDictType:
     return {"type": "image", "identifier": "".join([random.choice(string.ascii_uppercase) for _ in range(2)]), "uuid": str(uuid.uuid4())}
 
 
-def create_splitter_desc(orientation: str, splits, children):
-    return {"type": "splitter", "orientation": orientation, "splits": copy.copy(splits), "children": children}
+def create_splitter_desc(orientation: str, splits: typing.Sequence[float], children: typing.Sequence[Persistence.PersistentDictType]) -> Persistence.PersistentDictType:
+    return {"type": "splitter", "orientation": orientation, "splits": list(splits), "children": list(children)}
 
 
 class CreateWorkspaceCommand(Undo.UndoableCommand):
-    def __init__(self, workspace_controller: Workspace, name: str):
+    def __init__(self, workspace_controller: Workspace, name: str) -> None:
         super().__init__("Create Workspace")
         self.__workspace_controller = workspace_controller
         self.__workspace_layout_uuid = workspace_controller._workspace.uuid
         self.__new_name = name
-        self.__new_layout = None
-        self.__new_workspace_id = None
+        self.__new_layout: typing.Optional[Persistence.PersistentDictType] = None
+        self.__new_workspace_id: typing.Optional[str] = None
         self.initialize()
 
     def _get_modified_state(self) -> typing.Any:
@@ -136,7 +139,7 @@ class CloneWorkspaceCommand(Undo.UndoableCommand):
         self.__workspace_layout_uuid = workspace_controller._workspace.uuid
         self.__new_name = name
         self.__new_layout = workspace_controller._workspace.layout
-        self.__new_workspace_id = None
+        self.__new_workspace_id: typing.Optional[str] = None
         self.initialize()
 
     def _get_modified_state(self) -> typing.Any:
@@ -163,15 +166,15 @@ class CloneWorkspaceCommand(Undo.UndoableCommand):
 
 
 class ChangeWorkspaceContentsCommand(Undo.UndoableCommand):
-    def __init__(self, workspace_controller: Workspace, title: str, old_workspace_layout: dict = None):
+    def __init__(self, workspace_controller: Workspace, title: str, old_workspace_layout: typing.Optional[Persistence.PersistentDictType] = None):
         super().__init__(title)
         self.__workspace_controller = workspace_controller
         self.__old_workspace_layout = old_workspace_layout if old_workspace_layout else workspace_controller.deconstruct()
-        self.__new_workspace_layout = typing.cast(dict, None)
+        self.__new_workspace_layout: typing.Optional[Persistence.PersistentDictType] = None
         self.initialize()
 
     @property
-    def _old_workspace_layout(self):
+    def _old_workspace_layout(self) -> typing.Optional[Persistence.PersistentDictType]:
         return self.__old_workspace_layout
 
     def _get_modified_state(self) -> typing.Any:
@@ -185,6 +188,7 @@ class ChangeWorkspaceContentsCommand(Undo.UndoableCommand):
         self.__workspace_controller.reconstruct(self.__old_workspace_layout)
 
     def _redo(self) -> None:
+        assert self.__new_workspace_layout is not None
         self.__workspace_controller.reconstruct(self.__new_workspace_layout)
 
 
@@ -204,7 +208,7 @@ class Workspace:
     def __init__(self, document_controller: DocumentController.DocumentController, workspace_id: str):
         self.__document_controller_weakref = weakref.ref(document_controller)
 
-        self.ui = self.document_controller.ui
+        self.ui = document_controller.ui
 
         self.workspace_manager = WorkspaceManager()
 
@@ -213,7 +217,7 @@ class Workspace:
         self.dock_panels: typing.List[Panel.Panel] = []
         self.display_panels: typing.List[DisplayPanel.DisplayPanel] = []
 
-        self.__canvas_item = None
+        self.__canvas_item = typing.cast(CanvasItem.CanvasItemComposition, None)
 
         # create the root element
         root_widget = self.ui.create_column_widget(properties={"min-width": 640, "min-height": 480})
@@ -241,7 +245,7 @@ class Workspace:
 
         self.__workspace: typing.Optional[WorkspaceLayout.WorkspaceLayout] = None
         self.__change_splitter_command: typing.Optional[Undo.UndoableCommand] = None
-        self.__change_splitter_splits = None
+        self.__change_splitter_splits: typing.List[float] = list()
 
     def close(self) -> None:
         for message_box_widget in copy.copy(list(self.__message_boxes.values())):
@@ -250,13 +254,13 @@ class Workspace:
         if self.__workspace:
             self.__sync_layout()
         self.display_panels = []
-        self.__canvas_item = None
+        self.__canvas_item = typing.cast(CanvasItem.CanvasItemComposition, None)
         self.__workspace = None
         for dock_panel in self.dock_panels:
             dock_panel.close()
-        self.__content_column = None
-        self.filter_row = None
-        self.image_row = None
+        self.__content_column = typing.cast(typing.Any, None)
+        self.filter_row = typing.cast(typing.Any, None)
+        self.image_row = typing.cast(typing.Any, None)
         self.document_controller.detach_widget()
 
     @property
@@ -296,7 +300,7 @@ class Workspace:
         return self.document_controller.project
 
     @property
-    def _canvas_item(self) -> typing.Optional[CanvasItem.CanvasItemComposition]:
+    def _canvas_item(self) -> CanvasItem.CanvasItemComposition:
         return self.__canvas_item
 
     def _find_dock_panel(self, dock_panel_id: str) -> typing.Optional[Panel.Panel]:
@@ -321,7 +325,8 @@ class Workspace:
                         dock_panel.hide()
 
     def create_panel(self, document_controller: DocumentController.DocumentController, panel_id: str, title: str,
-                     positions: typing.Sequence[str], position: str, properties: dict) -> typing.Optional[Panel.Panel]:
+                     positions: typing.Sequence[str], position: str,
+                     properties: typing.Optional[Persistence.PersistentDictType]) -> typing.Optional[Panel.Panel]:
         try:
             panel = self.workspace_manager.create_panel_content(document_controller, panel_id, title, positions, position, properties)
             assert panel is not None, "panel is None [%s]" % panel_id
@@ -335,27 +340,29 @@ class Workspace:
             traceback.print_stack()
             return None
 
-    def display_display_item_in_display_panel(self, display_item, display_panel_id):
+    def display_display_item_in_display_panel(self, display_item: DisplayItem.DisplayItem, display_panel_id: str) -> None:
         for display_panel in self.display_panels:
             if display_panel.display_panel_id == display_panel_id:
                 display_panel.set_display_panel_display_item(display_item)
                 self.__sync_layout()
 
-    def _construct(self, desc: dict, display_panels: typing.List[DisplayPanel.DisplayPanel]) -> typing.Tuple[typing.Optional[CanvasItem.AbstractCanvasItem], typing.Optional[DisplayPanel.DisplayPanel]]:
+    def _construct(self, desc: Persistence.PersistentDictType, display_panels: typing.List[DisplayPanel.DisplayPanel]) -> typing.Tuple[typing.Optional[CanvasItem.AbstractCanvasItem], typing.Optional[DisplayPanel.DisplayPanel]]:
         selected_display_panel = None
         type = desc["type"]
         container = None
         item = None
-        post_children_adjust = lambda: None
+        def _post_children_adjust() -> None: pass
+        post_children_adjust = _post_children_adjust
         if type == "splitter":
-            container = CanvasItem.SplitterCanvasItem(orientation=desc.get("orientation"))
-            container.on_splits_will_change = functools.partial(self._splits_will_change, container)
-            container.on_splits_changed = functools.partial(self._splits_did_change, container)
-            def splitter_post_children_adjust():
+            splitter_canvas_item = CanvasItem.SplitterCanvasItem(orientation=desc.get("orientation"))
+            splitter_canvas_item.on_splits_will_change = functools.partial(self._splits_will_change, splitter_canvas_item)
+            splitter_canvas_item.on_splits_changed = functools.partial(self._splits_did_change, splitter_canvas_item)
+            def splitter_post_children_adjust() -> None:
                 splits = desc.get("splits")
                 if splits is not None:
-                    container.splits = splits
+                    splitter_canvas_item.splits = splits
             post_children_adjust = splitter_post_children_adjust
+            container = splitter_canvas_item
         elif type == "image":
             display_panel = DisplayPanel.DisplayPanel(self.document_controller, desc)
             display_panel.on_contents_changed = self.__sync_layout
@@ -367,28 +374,27 @@ class Workspace:
             children = desc.get("children", list())
             for child_desc in children:
                 child_canvas_item, child_selected_display_panel = self._construct(child_desc, display_panels)
-                container.add_canvas_item(child_canvas_item)
+                if child_canvas_item:
+                    container.add_canvas_item(child_canvas_item)
                 selected_display_panel = child_selected_display_panel if child_selected_display_panel else selected_display_panel
             post_children_adjust()
             return container, selected_display_panel
         return item, selected_display_panel
 
-    def deconstruct(self) -> dict:
-        assert self.__canvas_item
+    def deconstruct(self) -> Persistence.PersistentDictType:
         return self._deconstruct(self.__canvas_item.canvas_items[0])
 
-    def reconstruct(self, d: dict) -> None:
-        assert self.__canvas_item
-        display_panels = list()
+    def reconstruct(self, d: Persistence.PersistentDictType) -> None:
+        display_panels: typing.List[DisplayPanel.DisplayPanel] = list()
         selected_display_panel = self._reconstruct(d, self.__canvas_item, self.__canvas_item.canvas_items[0], display_panels)
         self.display_panels = display_panels
         self.document_controller.selected_display_panel = selected_display_panel
 
-    def _reconstruct(self, d: dict, container: CanvasItem.CanvasItemComposition, canvas_item: CanvasItem.AbstractCanvasItem, display_panels: typing.List[DisplayPanel.DisplayPanel]) -> typing.Optional[DisplayPanel.DisplayPanel]:
+    def _reconstruct(self, d: Persistence.PersistentDictType, container: CanvasItem.CanvasItemComposition, canvas_item: CanvasItem.AbstractCanvasItem, display_panels: typing.List[DisplayPanel.DisplayPanel]) -> typing.Optional[DisplayPanel.DisplayPanel]:
         selected_display_panel = None
         type = d["type"]
         if type == "splitter":
-            if isinstance(canvas_item, CanvasItem.SplitterCanvasItem):
+            if isinstance(canvas_item, CanvasItem.SplitterCanvasItem) and canvas_item.orientation == d.get("orientation", None):
                 canvas_item_container = canvas_item
                 children = d.get("children", list())
                 if len(children) == len(canvas_item_container.canvas_items):
@@ -402,7 +408,8 @@ class Workspace:
                     return selected_display_panel
             # fall through
             new_canvas_item, selected_display_panel = self._construct(d, display_panels)
-            container.replace_canvas_item(canvas_item, new_canvas_item)
+            if new_canvas_item:
+                container.replace_canvas_item(canvas_item, new_canvas_item)
             return selected_display_panel
         elif type == "image":
             if isinstance(canvas_item, DisplayPanel.DisplayPanel):
@@ -414,7 +421,8 @@ class Workspace:
                 return selected_display_panel
             # fall through
             new_canvas_item, selected_display_panel = self._construct(d, display_panels)
-            container.replace_canvas_item(canvas_item, new_canvas_item)
+            if new_canvas_item:
+                container.replace_canvas_item(canvas_item, new_canvas_item)
             return selected_display_panel
         return selected_display_panel
 
@@ -430,7 +438,7 @@ class Workspace:
                 return display_panel
         return None
 
-    def _deconstruct(self, canvas_item: CanvasItem.AbstractCanvasItem) -> dict:
+    def _deconstruct(self, canvas_item: CanvasItem.AbstractCanvasItem) -> Persistence.PersistentDictType:
         if isinstance(canvas_item, CanvasItem.SplitterCanvasItem):
             children = [self._deconstruct(child_canvas_item) for child_canvas_item in canvas_item.canvas_items]
             d = create_splitter_desc(canvas_item.orientation, canvas_item.splits, children)
@@ -450,7 +458,7 @@ class Workspace:
         return self.__workspace
 
     @property
-    def _workspace_layout(self) -> dict:
+    def _workspace_layout(self) -> Persistence.PersistentDictType:
         return self.deconstruct()
 
     def close_display_panels(self, display_panels: typing.Sequence[DisplayPanel.DisplayPanel]) -> None:
@@ -500,10 +508,10 @@ class Workspace:
             if display_panel != self.document_controller.selected_display_panel:
                 self.document_controller.add_secondary_display_panel(display_panel)
 
-    def switch_to_display_content(self, display_panel: DisplayPanel.DisplayPanel, display_panel_type, display_item: DisplayItem.DisplayItem = None) -> None:
-        d = {"type": "image", "display-panel-type": display_panel_type}
+    def switch_to_display_content(self, display_panel: DisplayPanel.DisplayPanel, display_panel_type: str, display_item: typing.Optional[DisplayItem.DisplayItem] = None) -> None:
+        d: Persistence.PersistentDictType = {"type": "image", "display-panel-type": display_panel_type}
         if display_item and display_panel_type != "empty-display-panel":
-            d["display_item_specifier"] = display_item.project.create_specifier(display_item).write()
+            d["display_item_specifier"] = Persistence.write_persistent_specifier(display_item.uuid)
         command = ChangeWorkspaceContentsCommand(self, _("Replace Display Panel"))
         display_panel.change_display_panel_content(d)
         self.document_controller.push_undo_command(command)
@@ -536,7 +544,8 @@ class Workspace:
         display_panels: typing.List[DisplayPanel.DisplayPanel]
         try:
             display_panels = list()  # to be populated by _construct
-            canvas_item, selected_display_panel = self._construct(workspace_layout.layout, display_panels)
+            workspace_layout_d = workspace_layout.layout or dict()
+            canvas_item, selected_display_panel = self._construct(workspace_layout_d, display_panels)
             # store the new workspace
             if canvas_item:
                 self.__workspace = workspace_layout
@@ -565,7 +574,7 @@ class Workspace:
         self.document_controller.project.workspace_uuid = workspace_layout.uuid
         self.document_controller._workspace_changed(workspace_layout)
 
-    def restore(self, workspace_uuid: uuid.UUID) -> None:
+    def restore(self, workspace_uuid: typing.Optional[uuid.UUID]) -> None:
         """
             Restore the workspace to the given workspace_uuid.
 
@@ -590,7 +599,7 @@ class Workspace:
         workspace_index = (workspace_index + 1) % len(self._project.workspaces)
         self.change_workspace(self._project.workspaces[workspace_index])
 
-    def new_workspace(self, name: typing.Optional[str] = None, layout: typing.Optional[dict] = None,
+    def new_workspace(self, name: typing.Optional[str] = None, layout: typing.Optional[Persistence.PersistentDictType] = None,
                       workspace_id: typing.Optional[str] = None,
                       index: typing.Optional[int] = None) -> WorkspaceLayout.WorkspaceLayout:
         """ Create a new workspace, insert into document_model, and return it. """
@@ -604,7 +613,7 @@ class Workspace:
             workspace.workspace_id = workspace_id
         return workspace
 
-    def ensure_workspace(self, name: str, layout: dict, workspace_id: str) -> None:
+    def ensure_workspace(self, name: str, layout: Persistence.PersistentDictType, workspace_id: str) -> None:
         """Looks for a workspace with workspace_id.
 
         If none is found, create a new one, add it, and change to it.
@@ -624,7 +633,7 @@ class Workspace:
                                     rejected_fn: typing.Optional[typing.Callable[[], None]] = None,
                                     accepted_text: typing.Optional[str] = None,
                                     rejected_text: typing.Optional[str] = None,
-                                    message_box_id: typing.Optional[str] = None) -> typing.Optional[UserInterface.Widget]:
+                                    message_box_id: typing.Optional[str] = None) -> typing.Optional[UserInterface.BoxWidget]:
         message_box_id = message_box_id if message_box_id else str(uuid.uuid4())
         if message_box_id in self.__message_boxes:
             return None
@@ -637,30 +646,40 @@ class Workspace:
         caption_row.add_stretch()
         inside_row = self.ui.create_row_widget()
 
-        def reject_button_clicked():
+        def reject_button_clicked() -> bool:
             if rejected_fn: rejected_fn()
-            async def remove_widget():
+            async def remove_widget() -> None:
                 self.message_column.remove(message_box_widget)
             self.document_controller.event_loop.create_task(remove_widget())
+            assert message_box_id
             del self.__message_boxes[message_box_id]
             return False
 
-        def accept_button_clicked():
-            accepted_fn(string_edit_widget.text)
-            async def remove_widget():
+        def accept_button_clicked() -> bool:
+            accepted_fn(string_edit_widget.text or str())
+            async def remove_widget() -> None:
                 self.message_column.remove(message_box_widget)
             self.document_controller.event_loop.create_task(remove_widget())
+            assert message_box_id
             del self.__message_boxes[message_box_id]
             return False
 
+        # dummy to pass typing
+        def reject_button_clicked_() -> None:
+            reject_button_clicked()
+
+        # dummy to pass typing
+        def accept_button_clicked_() -> None:
+            accept_button_clicked()
+
         string_edit_widget = self.ui.create_line_edit_widget()
         string_edit_widget.text = text
         string_edit_widget.on_return_pressed = accept_button_clicked
         string_edit_widget.on_escape_pressed = reject_button_clicked
         reject_button = self.ui.create_push_button_widget(rejected_text)
-        reject_button.on_clicked = reject_button_clicked
+        reject_button.on_clicked = reject_button_clicked_
         accepted_button = self.ui.create_push_button_widget(accepted_text)
-        accepted_button.on_clicked = accept_button_clicked
+        accepted_button.on_clicked = accept_button_clicked_
         inside_row.add_spacing(12)
         inside_row.add(string_edit_widget)
         inside_row.add_spacing(12)
@@ -683,7 +702,7 @@ class Workspace:
                                       rejected_fn: typing.Optional[typing.Callable[[], None]] = None,
                                       accepted_text: typing.Optional[str] = None,
                                       rejected_text: typing.Optional[str] = None, display_rejected: bool = True,
-                                      message_box_id: typing.Optional[str] = None) -> typing.Optional[UserInterface.Widget]:
+                                      message_box_id: typing.Optional[str] = None) -> typing.Optional[UserInterface.BoxWidget]:
         message_box_id = message_box_id if message_box_id else str(uuid.uuid4())
         if message_box_id in self.__message_boxes:
             return None
@@ -691,14 +710,16 @@ class Workspace:
         if rejected_text is None: rejected_text = _("Cancel")
         message_box_widget = self.ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}
 
-        def reject_button_clicked():
+        def reject_button_clicked() -> None:
             if rejected_fn: rejected_fn()
             self.message_column.remove(message_box_widget)
+            assert message_box_id
             del self.__message_boxes[message_box_id]
 
-        def accept_button_clicked():
+        def accept_button_clicked() -> None:
             accepted_fn()
             self.message_column.remove(message_box_widget)
+            assert message_box_id
             del self.__message_boxes[message_box_id]
 
         reject_button = self.ui.create_push_button_widget(rejected_text)
@@ -721,25 +742,24 @@ class Workspace:
         self.__message_boxes[message_box_id] = message_box_widget
         return message_box_widget
 
-    def pose_tool_tip_box(self, caption: str, timeout: float, message_box_id: typing.Optional[str]=None) -> typing.Optional[UserInterface.Widget]:
-        import threading
-        import time
+    def pose_tool_tip_box(self, caption: str, timeout: float, message_box_id: typing.Optional[str] = None) -> typing.Optional[UserInterface.BoxWidget]:
         message_box_id = message_box_id if message_box_id else str(uuid.uuid4())
         if message_box_id in self.__message_boxes:
             return None
         accepted_text = '\u274C'
         message_box_widget = self.ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}
         lock = threading.Lock()
-        def remove_box():
+        def remove_box() -> None:
             with lock:
                 if message_box_id in self.__message_boxes:
                     self.message_column.remove(message_box_widget)
+                    assert message_box_id
                     del self.__message_boxes[message_box_id]
 
-        def accept_button_clicked():
+        def accept_button_clicked() -> None:
             remove_box()
 
-        def wait_for_timeout():
+        def wait_for_timeout() -> None:
             time.sleep(timeout)
             self.document_controller.queue_task(remove_box)
 
@@ -757,7 +777,7 @@ class Workspace:
         self.message_column.add(message_box_widget)
         self.__message_boxes[message_box_id] = message_box_widget
         threading.Thread(target=wait_for_timeout, daemon=True).start()
-        message_box_widget.remove_now = remove_box
+        setattr(message_box_widget, "remove_now", remove_box)  # argh. type checking.
         return message_box_widget
 
     def handle_drag_enter(self, display_panel: DisplayPanel.DisplayPanel, mime_data: UserInterface.MimeData) -> str:
@@ -769,8 +789,8 @@ class Workspace:
             return "copy"
         return "ignore"
 
-    def handle_drag_leave(self, display_panel: DisplayPanel.DisplayPanel) -> bool:
-        return False
+    def handle_drag_leave(self, display_panel: DisplayPanel.DisplayPanel) -> str:
+        return "ignore"
 
     def handle_drag_move(self, display_panel: DisplayPanel.DisplayPanel, mime_data: UserInterface.MimeData, x: int, y: int) -> str:
         if mime_data.has_format(MimeTypes.DISPLAY_ITEM_MIME_TYPE):
@@ -784,7 +804,7 @@ class Workspace:
     def should_handle_drag_for_mime_data(self, mime_data: UserInterface.MimeData) -> bool:
         return mime_data.has_format(MimeTypes.DISPLAY_ITEM_MIME_TYPE) or mime_data.has_format("text/uri-list") or mime_data.has_format(MimeTypes.DISPLAY_PANEL_MIME_TYPE)
 
-    def handle_drop(self, display_panel: DisplayPanel.DisplayPanel, mime_data: UserInterface.MimeData, region, x: int, y: int) -> str:
+    def handle_drop(self, display_panel: DisplayPanel.DisplayPanel, mime_data: UserInterface.MimeData, region: str, x: int, y: int) -> str:
         document_model = self.document_model
         if mime_data.has_format(MimeTypes.DISPLAY_PANEL_MIME_TYPE):
             display_item, d = MimeTypes.mime_data_get_panel(mime_data, self.document_model)
@@ -816,14 +836,15 @@ class Workspace:
 
     def _replace_displayed_display_item(self, display_panel: DisplayPanel.DisplayPanel,
                                         display_item: typing.Optional[DisplayItem.DisplayItem],
-                                        d: typing.Optional[dict] = None) -> Undo.UndoableCommand:
+                                        d: typing.Optional[Persistence.PersistentDictType] = None) -> Undo.UndoableCommand:
         return self.__replace_displayed_display_item(display_panel, display_item, d)
 
     def __replace_displayed_display_item(self, display_panel: DisplayPanel.DisplayPanel,
                                          display_item: typing.Optional[DisplayItem.DisplayItem],
-                                         d: typing.Optional[dict] = None) -> Undo.UndoableCommand:
+                                         d: typing.Optional[Persistence.PersistentDictType] = None) -> Undo.UndoableCommand:
         """ Used in drag/drop support. """
-        self.document_controller.replaced_display_panel_content = display_panel.save_contents()
+        if self.document_controller.replaced_display_panel_content_flag:
+            self.document_controller.replaced_display_panel_content = display_panel.save_contents()
         command = ChangeWorkspaceContentsCommand(self, _("Replace Display Panel"))
         if display_item:
             display_panel.set_display_panel_display_item(display_item, detect_controller=True)
@@ -835,21 +856,22 @@ class Workspace:
 
     def insert_display_panel(self, display_panel: DisplayPanel.DisplayPanel, region: str,
                              display_item: typing.Optional[DisplayItem.DisplayItem] = None,
-                             d: typing.Optional[dict] = None, new_uuid: typing.Optional[uuid.UUID] = None,
-                             new_splits: typing.Optional[typing.List[int]] = None) -> Undo.UndoableCommand:
+                             d: typing.Optional[Persistence.PersistentDictType] = None, new_uuid: typing.Optional[uuid.UUID] = None,
+                             new_splits: typing.Optional[typing.List[float]] = None) -> Undo.UndoableCommand:
         assert self.__workspace
         command = ChangeWorkspaceContentsCommand(self, _("Split Display Panel"))
         self._insert_display_panel(display_panel, region, display_item, d, new_uuid, new_splits)
         return command
 
     def _insert_display_panel(self, display_panel: DisplayPanel.DisplayPanel, region: str,
-                              display_item: typing.Optional[DisplayItem.DisplayItem], d: typing.Optional[dict],
+                              display_item: typing.Optional[DisplayItem.DisplayItem], d: typing.Optional[Persistence.PersistentDictType],
                               new_uuid: typing.Optional[uuid.UUID],
-                              new_splits: typing.Optional[typing.List[int]] = None) -> typing.Tuple[typing.Optional[typing.List[int]], typing.Optional[DisplayPanel.DisplayPanel]]:
+                              new_splits: typing.Optional[typing.List[float]] = None) -> typing.Tuple[typing.Optional[typing.List[float]], typing.Optional[DisplayPanel.DisplayPanel]]:
         assert isinstance(display_panel, DisplayPanel.DisplayPanel)
         orientation = "vertical" if region in ("left", "right") else "horizontal"
         new_display_panel = None
         container = display_panel.container
+        assert container
         # record old splits for undo if modifying an existing splitter; otherwise removing the display panel is trivial
         old_splits = list(container.splits) if isinstance(container, CanvasItem.SplitterCanvasItem) and container.orientation == orientation else None
         # if not modifying an existing splitter or if modifying in the other orientation, wrap panel in new splitter
@@ -864,7 +886,7 @@ class Workspace:
         index = container.canvas_items.index(display_panel)
         if isinstance(container, CanvasItem.SplitterCanvasItem):
             # modify the existing splitter
-            old_split = container.splits[index] if not new_splits else None
+            old_split = container.splits[index] if not new_splits else 0.0
             new_index_adj = 1 if region == "right" or region == "bottom" else 0
             new_display_panel = DisplayPanel.DisplayPanel(self.document_controller, dict(), new_uuid)
             self.display_panels.insert(self.display_panels.index(display_panel) + new_index_adj, new_display_panel)
@@ -887,32 +909,32 @@ class Workspace:
         return old_splits, new_display_panel
 
     def remove_display_panel(self, display_panel: DisplayPanel.DisplayPanel,
-                             splits: typing.Optional[typing.List[int]] = None) -> Undo.UndoableCommand:
+                             splits: typing.Optional[typing.List[float]] = None) -> Undo.UndoableCommand:
         assert self.__workspace
         command = ChangeWorkspaceContentsCommand(self, _("Remove Display Panel"))
         self._remove_display_panel(display_panel, splits)
         return command
 
     def _remove_display_panel(self, display_panel: DisplayPanel.DisplayPanel,
-                              splits: typing.Optional[typing.List[int]]) -> typing.Tuple[typing.Optional[DisplayPanel.DisplayPanel], typing.Optional[typing.List[int]], str]:
+                              splits: typing.Optional[typing.List[float]]) -> typing.Tuple[typing.Optional[DisplayPanel.DisplayPanel], typing.Optional[typing.List[float]], str]:
         # first make sure the display panel has no content
         display_panel.change_display_panel_content({"type": "image", "display-panel-type": "empty-display-panel"})
         # now remove it
         container = display_panel.container
         region_id = str()
-        old_display_panel = None
-        old_splits = None
+        old_display_panel: typing.Optional[DisplayPanel.DisplayPanel] = None
+        old_splits: typing.Optional[typing.List[float]] = None
         if isinstance(container, CanvasItem.SplitterCanvasItem):
-            old_splits = container.splits
+            old_splits = list(container.splits)
             if display_panel in container.canvas_items:
                 if len(container.canvas_items) > 1:
                     # configure the redo
                     display_panel_index = container.canvas_items.index(display_panel)
                     if display_panel_index > 0:
-                        old_display_panel = container.canvas_items[display_panel_index - 1]
+                        old_display_panel = typing.cast(DisplayPanel.DisplayPanel, container.canvas_items[display_panel_index - 1])
                         region_id = "right" if container.orientation == "vertical" else "bottom"
                     else:
-                        old_display_panel = container.canvas_items[1]
+                        old_display_panel = typing.cast(DisplayPanel.DisplayPanel, container.canvas_items[1])
                         region_id = "left" if container.orientation == "vertical" else "top"
                 self.display_panels.remove(display_panel)
                 container.remove_canvas_item(display_panel)
@@ -925,7 +947,8 @@ class Workspace:
         return old_display_panel, old_splits, region_id
 
     def apply_layout(self, display_panel: DisplayPanel.DisplayPanel, w: int, h: int) -> typing.List[DisplayPanel.DisplayPanel]:
-        assert isinstance(display_panel, DisplayPanel.DisplayPanel)
+        display_panel_container = display_panel.container
+        assert display_panel_container
         assert self.__workspace
         change_workspace_contents_command = ChangeWorkspaceContentsCommand(self, _("Change Workspace Contents"))
 
@@ -935,7 +958,7 @@ class Workspace:
         row_splitter_canvas_item = CanvasItem.SplitterCanvasItem(orientation="horizontal")
         row_splitter_canvas_item.on_splits_will_change = functools.partial(self._splits_will_change, row_splitter_canvas_item)
         row_splitter_canvas_item.on_splits_changed = functools.partial(self._splits_did_change, row_splitter_canvas_item)
-        display_panel.container.wrap_canvas_item(display_panel, row_splitter_canvas_item)
+        display_panel_container.wrap_canvas_item(display_panel, row_splitter_canvas_item)
         new_display_panels.append(display_panel)
         dest_index = self.display_panels.index(display_panel)
         for row in range(h):
@@ -952,16 +975,18 @@ class Workspace:
             column_splitter_canvas_item = CanvasItem.SplitterCanvasItem(orientation="vertical")
             column_splitter_canvas_item.on_splits_will_change = functools.partial(self._splits_will_change, column_splitter_canvas_item)
             column_splitter_canvas_item.on_splits_changed = functools.partial(self._splits_did_change, column_splitter_canvas_item)
-            row_display_panel.container.wrap_canvas_item(row_display_panel, column_splitter_canvas_item)
+            row_display_panel_container = row_display_panel.container
+            assert row_display_panel_container
+            row_display_panel_container.wrap_canvas_item(row_display_panel, column_splitter_canvas_item)
             for column in range(1, w):
                 column_display_panel = DisplayPanel.DisplayPanel(self.document_controller, dict())
                 new_display_panels.append(column_display_panel)
                 self.display_panels.insert(dest_index + 1, column_display_panel)
                 dest_index += 1
                 column_splitter_canvas_item.insert_canvas_item(column + 1, column_display_panel)
-            column_splitter_canvas_item.splits = [1/w] * w
+            column_splitter_canvas_item.splits = [1//w] * w
 
-        row_splitter_canvas_item.splits = [1/h] * h
+        row_splitter_canvas_item.splits = [1//h] * h
 
         self.__sync_layout()
         display_panel.request_focus()
@@ -971,7 +996,7 @@ class Workspace:
         return new_display_panels
 
     def _splits_will_change(self, splitter_canvas_item: CanvasItem.SplitterCanvasItem) -> None:
-        self.__change_splitter_splits = splitter_canvas_item.splits
+        self.__change_splitter_splits = list(splitter_canvas_item.splits)
         self.__change_splitter_command = ChangeWorkspaceContentsCommand(self, _("Change Workspace Contents"))
 
     def _splits_did_change(self, splitter_canvas_item: CanvasItem.SplitterCanvasItem) -> None:
@@ -982,7 +1007,7 @@ class Workspace:
             else:
                 self.__change_splitter_command.close()
             self.__change_splitter_command = None
-            self.__change_splitter_splits = None
+            self.__change_splitter_splits = list()
 
     def _sync_layout(self) -> None:
         self.__sync_layout()
@@ -990,7 +1015,6 @@ class Workspace:
     def __sync_layout(self) -> None:
         # ensure that the layout is written to persistent storage
         assert self.__workspace
-        assert self.__canvas_item
         self.__workspace.layout = self._deconstruct(self.__canvas_item.canvas_items[0])
 
 
@@ -1000,18 +1024,21 @@ class WorkspaceManager(metaclass=Utility.Singleton):
         The WorkspaceManager object keeps a list of workspaces and a list of panel
         types. It also creates workspace objects.
         """
-    def __init__(self):
-        self.__panel_tuples = {}
+    def __init__(self) -> None:
+        self.__panel_tuples: typing.Dict[str, typing.Tuple[typing.Type[typing.Any], str, str, typing.List[str], str, typing.Optional[Persistence.PersistentDictType]]] = dict()
 
-    def register_panel(self, panel_class, panel_id: str, name: str, positions: typing.List[str], position: str, properties: typing.Optional[dict]=None) -> None:
-        panel_tuple = panel_class, panel_id, name, positions, position, properties
+    def register_panel(self, panel_class: typing.Type[typing.Any], panel_id: str, name: str,
+                       positions: typing.Sequence[str], position: str,
+                       properties: typing.Optional[Persistence.PersistentDictType] = None) -> None:
+        panel_tuple = panel_class, panel_id, name, list(positions), position, properties
         self.__panel_tuples[panel_id] = panel_tuple
 
     def unregister_panel(self, panel_id: str) -> None:
         del self.__panel_tuples[panel_id]
 
     def create_panel_content(self, document_controller: DocumentController.DocumentController, panel_id: str,
-                             title: str, positions: typing.Sequence[str], position: str, properties: dict) -> typing.Optional[Panel.Panel]:
+                             title: str, positions: typing.Sequence[str], position: str,
+                             properties: typing.Optional[Persistence.PersistentDictType]) -> typing.Optional[Panel.Panel]:
         if panel_id in self.__panel_tuples:
             tuple = self.__panel_tuples[panel_id]
             cls = tuple[0]
@@ -1019,7 +1046,7 @@ class WorkspaceManager(metaclass=Utility.Singleton):
                 properties = properties if properties else {}
                 panel = cls(document_controller, panel_id, properties)
                 panel.create_dock_widget(title, positions, position)
-                return panel
+                return typing.cast(Panel.Panel, panel)
             except Exception as e:
                 import traceback
                 print("Exception creating panel '" + panel_id + "': " + str(e))
@@ -1034,7 +1061,7 @@ class WorkspaceManager(metaclass=Utility.Singleton):
         assert self.__filter_panel_class
         return self.__filter_panel_class(document_controller)
 
-    def get_panel_info(self, panel_id: str) -> typing.Tuple:
+    def get_panel_info(self, panel_id: str) -> typing.Tuple[str, typing.Sequence[str], str, typing.Optional[Persistence.PersistentDictType]]:
         assert panel_id in self.__panel_tuples
         tuple = self.__panel_tuples[panel_id]
         return tuple[2], tuple[3], tuple[4], tuple[5]
