@@ -8,6 +8,7 @@ import json
 import logging
 import os
 import pathlib
+import types
 import typing
 import uuid
 
@@ -22,38 +23,45 @@ from nion.swift.model import Persistence
 from nion.swift.model import Project
 from nion.swift.model import Schema
 from nion.utils import Converter
-from nion.utils import Observable
-
 
 _ = gettext.gettext
 
-
 ProfileContext = typing.TypeVar("ProfileContext")
 
 
-class ProjectReference(Observable.Observable, Persistence.PersistentObject):
+class ProjectReference(Persistence.PersistentObject):
 
-    def __init__(self, type: str):
+    def __init__(self, type: str) -> None:
         super().__init__()
         self.define_type(type)
-        self.define_property("project_uuid", converter=Converter.UuidToStringConverter())
-        self.define_property("last_used", None, converter=Converter.DatetimeToStringConverter())
+        self.define_property("project_uuid", converter=Converter.UuidToStringConverter(), hidden=True)
+        self.define_property("last_used", None, converter=Converter.DatetimeToStringConverter(), hidden=True)
         self.__has_project_info_been_read = False
         self.__project_version: typing.Optional[int] = None
         self.__project_state = "invalid"
         self.__document_model: typing.Optional[DocumentModel.DocumentModel] = None
-        self.__document_model_about_to_close_listener = None
-        self.storage_cache = None
-        self.project_uuid: typing.Optional[uuid.UUID] = None  # satisfy type checker
-        self.last_used: typing.Optional[datetime.datetime] = None  # satisfy type checker
 
-    def close(self) -> None:
-        if self.__document_model_about_to_close_listener:
-            self.__document_model_about_to_close_listener.close()
-            self.__document_model_about_to_close_listener = None
-        super().close()
+    @property
+    def project_uuid(self) -> typing.Optional[uuid.UUID]:
+        return typing.cast(typing.Optional[uuid.UUID], self._get_persistent_property_value("project_uuid"))
 
-    def about_to_be_removed(self, container):
+    @project_uuid.setter
+    def project_uuid(self, value: typing.Optional[uuid.UUID]) -> None:
+        self._set_persistent_property_value("project_uuid", value)
+
+    @property
+    def is_valid(self) -> bool:
+        return False
+
+    @property
+    def last_used(self) -> typing.Optional[datetime.datetime]:
+        return typing.cast(typing.Optional[datetime.datetime], self._get_persistent_property_value("last_used"))
+
+    @last_used.setter
+    def last_used(self, value: typing.Optional[datetime.datetime]) -> None:
+        self._set_persistent_property_value("last_used", value)
+
+    def about_to_be_removed(self, container: Persistence.PersistentObject) -> None:
         self.unload_project()
         super().about_to_be_removed(container)
 
@@ -61,30 +69,7 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
         self.unload_project()
         super().close_relationships()
 
-    def insert_model_item(self, container, name, before_index, item):
-        """Insert a model item. Let this item's container do it if possible; otherwise do it directly.
-
-        Passing responsibility to this item's container allows the library to easily track dependencies.
-        However, if this item isn't yet in the library hierarchy, then do the operation directly.
-        """
-        if self.container:
-            self.container.insert_model_item(container, name, before_index, item)
-        else:
-            container.insert_item(name, before_index, item)
-
-    def remove_model_item(self, container, name, item, *, safe: bool = False) -> Changes.UndeleteLog:
-        """Remove a model item. Let this item's container do it if possible; otherwise do it directly.
-
-        Passing responsibility to this item's container allows the library to easily track dependencies.
-        However, if this item isn't yet in the library hierarchy, then do the operation directly.
-        """
-        if self.container:
-            return self.container.remove_model_item(container, name, item, safe=safe)
-        else:
-            container.remove_item(name, item)
-            return Changes.UndeleteLog()
-
-    def read_from_dict(self, properties):
+    def read_from_dict(self, properties: Persistence.PersistentDictType) -> None:
         super().read_from_dict(properties)
         self.establish_last_used()
 
@@ -96,7 +81,7 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
         # fallback when last_used is not initialized
         return self.modified
 
-    def __property_changed(self, name, value):
+    def __property_changed(self, name: str, value: typing.Any) -> None:
         self.notify_property_changed(name)
 
     @property
@@ -132,7 +117,7 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
         return pathlib.Path()
 
     @property
-    def project_reference_parts(self) -> typing.Tuple[str]:
+    def project_reference_parts(self) -> typing.Sequence[str]:
         raise NotImplementedError()
 
     def make_storage(self, profile_context: typing.Optional[ProfileContext]) -> typing.Optional[FileStorageSystem.ProjectStorageSystem]:
@@ -142,10 +127,11 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
         if not self.__has_project_info_been_read:
             try:
                 project_storage_system = self.make_storage(profile_context)
+                if project_storage_system:
+                    project_storage_system.load_properties()
             except Exception:
                 project_storage_system = None
             if project_storage_system:
-                project_storage_system.load_properties()
                 with contextlib.closing(Project.Project(project_storage_system)) as project:
                     if self.project_uuid != project.project_uuid:
                         self.project_uuid = project.project_uuid
@@ -153,7 +139,7 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
                     self.__project_state = project.project_state
                     self.__has_project_info_been_read = True
 
-    def load_project(self, profile_context: typing.Optional[ProfileContext]) -> None:
+    def load_project(self, profile_context: typing.Optional[ProfileContext], cache_dir_path: typing.Optional[pathlib.Path], *, cache_factory: typing.Optional[Cache.CacheFactory]) -> None:
         """Read project.
 
         The profile context is used during testing.
@@ -165,16 +151,23 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
             project_storage_system = self.make_storage(profile_context)
             if project_storage_system:
                 project_storage_system.load_properties()
-                project = Project.Project(project_storage_system)
+                if not cache_factory:
+                    assert cache_dir_path
+
+                    def encode(uuid_: uuid.UUID, alphabet: str) -> str:
+                        result = str()
+                        uuid_int = uuid_.int
+                        while uuid_int:
+                            uuid_int, digit = divmod(uuid_int, len(alphabet))
+                            result += alphabet[digit]
+                        return result
+
+                    encoded_base_path = "proj_" + encode(self.project_uuid or self.uuid, "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890")  # 25 character results
+                    cache_factory = Cache.DbCacheFactory(cache_dir_path, str(encoded_base_path))
+                project = Project.Project(project_storage_system, cache_factory)
 
             if project:
-                self.__document_model = DocumentModel.DocumentModel(project, storage_cache=self.storage_cache)
-
-                # handle special case of document model closing during tests
-                def document_window_close():
-                    self.__document_model_about_to_close_listener.close()
-                    self.__document_model_about_to_close_listener = None
-                    self.__document_model = None
+                self.__document_model = DocumentModel.DocumentModel(project)
 
                 project.prepare_read_project()  # sets up the uuid, used next.
 
@@ -182,8 +175,6 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
                 project.about_to_be_inserted(self)
                 self.notify_property_changed("project")  # before reading, so document model has a chance to set up
                 project.read_project()
-
-                self.__document_model_about_to_close_listener = self.__document_model.about_to_close_event.listen(document_window_close)
             else:
                 logging.getLogger("loader").warning(f"Project could not be loaded {self}.")
 
@@ -194,7 +185,9 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
             self.project.about_to_be_removed(self)
             self.project.persistent_object_context = typing.cast(Persistence.PersistentObjectContext, None)
             assert self.__document_model
-            self.__document_model.close()
+            # add/remove ref in case it was never added anywhere else. this will delete it.
+            with self.__document_model.ref():
+                pass
             self.__document_model = None
             self.notify_property_changed("project")
 
@@ -222,21 +215,42 @@ class ProjectReference(Observable.Observable, Persistence.PersistentObject):
 class IndexProjectReference(ProjectReference):
     type = "project_index"
 
-    def __init__(self):
+    def __init__(self) -> None:
         super().__init__(self.__class__.type)
-        self.define_property("project_path", converter=Converter.PathToStringConverter())
+        self.define_property("project_path", converter=Converter.PathToStringConverter(), hidden=True)
 
     @property
-    def project_reference_parts(self) -> typing.Tuple[str]:
-        return self.project_path.parts if self.project_path else tuple()
+    def is_valid(self) -> bool:
+        project_path = self.project_path
+        return project_path is not None and project_path.exists()
+
+    @property
+    def project_path(self) -> typing.Optional[pathlib.Path]:
+        return typing.cast(typing.Optional[pathlib.Path], self._get_persistent_property_value("project_path"))
+
+    @project_path.setter
+    def project_path(self, value: typing.Optional[pathlib.Path]) -> None:
+        self._set_persistent_property_value("project_path", value)
+
+    @property
+    def project_reference_parts(self) -> typing.Sequence[str]:
+        project_path = self.project_path
+        return project_path.parts if project_path else tuple()
 
     def make_storage(self, profile_context: typing.Optional[ProfileContext]) -> typing.Optional[FileStorageSystem.ProjectStorageSystem]:
-        return FileStorageSystem.make_index_project_storage_system(self.project_path)
+        project_path = self.project_path
+        if project_path:
+            return FileStorageSystem.make_index_project_storage_system(project_path)
+        return None
 
     def _get_last_used(self) -> datetime.datetime:
         # fallback when last_used is not initialized
         try:
-            return datetime.datetime.fromtimestamp(os.path.getmtime(self.project_path))
+            project_path = self.project_path
+            if project_path:
+                return datetime.datetime.fromtimestamp(os.path.getmtime(project_path))
+            else:
+                return super()._get_last_used()
         except Exception:
             return super()._get_last_used()
 
@@ -244,12 +258,29 @@ class IndexProjectReference(ProjectReference):
 class FolderProjectReference(ProjectReference):
     type = "project_folder"
 
-    def __init__(self):
+    def __init__(self) -> None:
         super().__init__(self.__class__.type)
-        self.define_property("project_folder_path", converter=Converter.PathToStringConverter())
+        self.define_property("project_folder_path", converter=Converter.PathToStringConverter(), hidden=True)
 
     @property
-    def project_reference_parts(self) -> typing.Tuple[str]:
+    def is_valid(self) -> bool:
+        project_folder_path = self.project_folder_path
+        if project_folder_path:
+            for project_file, project_dir in FileStorageSystem.FileProjectStorageSystem._get_migration_paths(project_folder_path):
+                if project_file.exists():
+                    return True
+        return False
+
+    @property
+    def project_folder_path(self) -> typing.Optional[pathlib.Path]:
+        return typing.cast(typing.Optional[pathlib.Path], self._get_persistent_property_value("project_folder_path"))
+
+    @project_folder_path.setter
+    def project_folder_path(self, value: typing.Optional[pathlib.Path]) -> None:
+        self._set_persistent_property_value("project_folder_path", value)
+
+    @property
+    def project_reference_parts(self) -> typing.Sequence[str]:
         return self.project_folder_path.parts if self.project_folder_path else tuple()
 
     def make_storage(self, profile_context: typing.Optional[ProfileContext]) -> typing.Optional[FileStorageSystem.ProjectStorageSystem]:
@@ -260,7 +291,11 @@ class FolderProjectReference(ProjectReference):
     def _get_last_used(self) -> datetime.datetime:
         # fallback when last_used is not initialized
         try:
-            return datetime.datetime.fromtimestamp(os.path.getmtime(self.project_folder_path))
+            project_folder_path = self.project_folder_path
+            if project_folder_path:
+                return datetime.datetime.fromtimestamp(os.path.getmtime(project_folder_path))
+            else:
+                return super()._get_last_used()
         except Exception:
             return super()._get_last_used()
 
@@ -285,7 +320,7 @@ class FolderProjectReference(ProjectReference):
         return new_project_reference
 
 
-project_reference_factory_hook = None
+project_reference_factory_hook: typing.Optional[typing.Callable[[str], typing.Optional[ProjectReference]]] = None
 
 
 def project_reference_factory(lookup_id: typing.Callable[[str], str]) -> typing.Optional[ProjectReference]:
@@ -300,13 +335,13 @@ def project_reference_factory(lookup_id: typing.Callable[[str], str]) -> typing.
 
 
 class ScriptItem(Schema.Entity):
-    def __init__(self, entity_type: Schema.EntityType):
-        super().__init__(**entity_type.entity_init_kwargs())
-        self.persistent_storage = None
+    def __init__(self, entity_type: Schema.EntityType, context: typing.Optional[Schema.EntityContext] = None) -> None:
+        super().__init__(entity_type, context)
+        self.persistent_storage: typing.Optional[Persistence.PersistentStorageInterface] = None
 
     @property
     def is_closed(self) -> bool:
-        return self._get_field_value("is_closed")
+        return typing.cast(bool, self._get_field_value("is_closed"))
 
     @is_closed.setter
     def is_closed(self, value: bool) -> None:
@@ -327,7 +362,7 @@ class ScriptItem(Schema.Entity):
 
 
 class FileScriptItem(ScriptItem):
-    def __init__(self, path: typing.Optional[pathlib.Path] = None):
+    def __init__(self, path: typing.Optional[pathlib.Path] = None) -> None:
         super().__init__(Model.FileScriptItem)
         if path:
             self.path = path
@@ -345,7 +380,7 @@ class FileScriptItem(ScriptItem):
 
 
 class FolderScriptItem(ScriptItem):
-    def __init__(self, folder_path: typing.Optional[pathlib.Path] = None, is_closed: bool = True):
+    def __init__(self, folder_path: typing.Optional[pathlib.Path] = None, is_closed: bool = True) -> None:
         super().__init__(Model.FolderScriptItem)
         if folder_path:
             self.folder_path = folder_path
@@ -363,6 +398,7 @@ class FolderScriptItem(ScriptItem):
         self._set_field_value("folder_path", value)
 
 
+# casting required to use entity in place of persistent object
 def script_item_factory(lookup_id: typing.Callable[[str], str]) -> typing.Optional[ScriptItem]:
     type = lookup_id("type")
     if type == Model.FileScriptItem.entity_id:
@@ -372,29 +408,37 @@ def script_item_factory(lookup_id: typing.Callable[[str], str]) -> typing.Option
     return None
 
 
-class Profile(Observable.Observable, Persistence.PersistentObject):
+class Profile(Persistence.PersistentObject):
     count = 0  # useful for detecting leaks in tests
 
-    def __init__(self, storage_system=None, storage_cache=None, *, profile_context: typing.Optional[ProfileContext] = None):
+    def __init__(self, storage_system: typing.Optional[FileStorageSystem.PersistentStorageSystem] = None,
+                 cache_dir_path: typing.Optional[pathlib.Path] = None, *,
+                 cache_factory: typing.Optional[Cache.CacheFactory] = None,
+                 profile_context: typing.Optional[ProfileContext] = None) -> None:
         super().__init__()
         self.__class__.count += 1
 
         self.define_root_context()
         self.define_type("profile")
-        self.define_property("last_project_reference", converter=Converter.UuidToStringConverter())
-        self.define_property("work_project_reference_uuid", converter=Converter.UuidToStringConverter())
-        self.define_property("closed_items", list())
-        self.define_property("script_items_updated", False, changed=self.__property_changed)
+        self.define_property("last_project_reference", converter=Converter.UuidToStringConverter(), hidden=True)
+        self.define_property("work_project_reference_uuid", converter=Converter.UuidToStringConverter(), hidden=True)
+        self.define_property("closed_items", list(), hidden=True)
+        self.define_property("script_items_updated", False, changed=self.__property_changed, hidden=True)
         self.define_relationship("project_references", project_reference_factory,
-                                 insert=self.__insert_project_reference, remove=self.__remove_project_reference)
-        self.define_relationship("script_items", script_item_factory)
+                                 insert=self.__insert_project_reference, remove=self.__remove_project_reference,
+                                 hidden=True)
+        self.define_relationship("script_items", typing.cast(
+            typing.Callable[[typing.Callable[[str], str]], typing.Optional[Persistence.PersistentObject]],
+            script_item_factory), hidden=True)
 
         self.storage_system = storage_system or FileStorageSystem.MemoryPersistentStorageSystem()
         self.storage_system.load_properties()
 
-        self.storage_cache: typing.Any = storage_cache or Cache.DictStorageCache()  # need to deallocate
         self.set_storage_system(self.storage_system)
 
+        self.__cache_dir_path = cache_dir_path
+        self.__cache_factory = cache_factory
+
         self.profile_context = None
 
         # helper object to produce the projects sequence
@@ -408,17 +452,56 @@ class Profile(Observable.Observable, Persistence.PersistentObject):
             self.add_project_memory()
 
     def close(self) -> None:
-        self.storage_cache.close()
-        self.storage_cache = None
         self.storage_system.close()
-        self.storage_system = None
-        self.__projects_observer.close()
+        self.storage_system = typing.cast(typing.Any, None)
+        if self.__projects_observer:
+            self.__projects_observer.close()
         self.__projects_observer = typing.cast(Observer.AbstractItemSource, None)
         self.profile_context = None
         self.__class__.count -= 1
         super().close()
 
-    def read_from_dict(self, properties):
+    @property
+    def last_project_reference(self) -> typing.Optional[uuid.UUID]:
+        return typing.cast(typing.Optional[uuid.UUID], self._get_persistent_property_value("last_project_reference"))
+
+    @last_project_reference.setter
+    def last_project_reference(self, value: typing.Optional[uuid.UUID]) -> None:
+        self._set_persistent_property_value("last_project_reference", value)
+
+    @property
+    def work_project_reference_uuid(self) -> typing.Optional[uuid.UUID]:
+        return typing.cast(typing.Optional[uuid.UUID], self._get_persistent_property_value("work_project_reference_uuid"))
+
+    @work_project_reference_uuid.setter
+    def work_project_reference_uuid(self, value: typing.Optional[uuid.UUID]) -> None:
+        self._set_persistent_property_value("work_project_reference_uuid", value)
+
+    @property
+    def closed_items(self) -> typing.Sequence[str]:
+        return typing.cast(typing.Sequence[str], self._get_persistent_property_value("closed_items"))
+
+    @closed_items.setter
+    def closed_items(self, value: typing.Sequence[str]) -> None:
+        self._set_persistent_property_value("closed_items", value)
+
+    @property
+    def script_items_updated(self) -> bool:
+        return typing.cast(bool, self._get_persistent_property_value("script_items_updated"))
+
+    @script_items_updated.setter
+    def script_items_updated(self, value: bool) -> None:
+        self._set_persistent_property_value("script_items_updated", value)
+
+    @property
+    def project_references(self) -> typing.Sequence[ProjectReference]:
+        return typing.cast(typing.Sequence[ProjectReference], self._get_relationship_values("project_references"))
+
+    @property
+    def script_items(self) -> typing.Sequence[ScriptItem]:
+        return typing.cast(typing.Sequence[ScriptItem], self._get_relationship_values("script_items"))
+
+    def read_from_dict(self, properties: Persistence.PersistentDictType) -> None:
         # cleanup from beta versions
         properties.pop("data_groups", None)
         properties.pop("data_item_references", None)
@@ -429,33 +512,32 @@ class Profile(Observable.Observable, Persistence.PersistentObject):
         properties.pop("workspaces", None)
         super().read_from_dict(properties)
 
-    def __property_changed(self, name, value):
+    def __property_changed(self, name: str, value: typing.Any) -> None:
         self.notify_property_changed(name)
 
     def __insert_project_reference(self, name: str, before_index: int, project_reference: ProjectReference) -> None:
-        project_reference.storage_cache = self.storage_cache
         project_reference.read_project_info(self.profile_context)
         self.notify_insert_item("project_references", project_reference, before_index)
 
     def __remove_project_reference(self, name: str, index: int, project_reference: ProjectReference) -> None:
-        project_reference.storage_cache = None
         self.notify_remove_item("project_references", project_reference, index)
 
     @property
-    def projects(self) -> typing.List[Project.Project]:
-        return typing.cast(typing.List[Project.Project], self.__projects_observer.item)
+    def projects(self) -> typing.Sequence[Project.Project]:
+        assert self.__projects_observer
+        return typing.cast(typing.Sequence[Project.Project], self.__projects_observer.item)
 
     def insert_script_item(self, before_index: int, script_item: ScriptItem) -> None:
         """Insert a script_item before the index, but do it through the container, so dependencies can be tracked."""
-        self.insert_model_item(self, "script_items", before_index, script_item)
+        self.insert_model_item(self, "script_items", before_index, typing.cast(Persistence.PersistentObject, script_item))
 
     def append_script_item(self, script_item: ScriptItem) -> None:
         """Append a script_item, but do it through the container, so dependencies can be tracked."""
-        self.insert_model_item(self, "script_items", self.item_count("script_items"), script_item)
+        self.insert_model_item(self, "script_items", self.item_count("script_items"), typing.cast(Persistence.PersistentObject, script_item))
 
     def remove_script_item(self, script_item: ScriptItem, *, safe: bool = False) -> Changes.UndeleteLog:
         """Remove a script_item, but do it through the container, so dependencies can be tracked."""
-        return self.remove_model_item(self, "script_items", script_item, safe=safe)
+        return self.remove_model_item(self, "script_items", typing.cast(Persistence.PersistentObject, script_item), safe=safe)
 
     def __insert_script_item(self, name: str, before_index: int, script_item: ScriptItem) -> None:
         self.notify_insert_item("script_items", script_item, before_index)
@@ -464,52 +546,35 @@ class Profile(Observable.Observable, Persistence.PersistentObject):
         self.notify_remove_item("script_items", script_item, index)
 
     @property
-    def _profile_storage_system(self) -> FileStorageSystem.PersistentStorageSystem:
+    def _profile_storage_system(self) -> typing.Optional[FileStorageSystem.PersistentStorageSystem]:
         return self.storage_system
 
-    def insert_model_item(self, container, name, before_index, item):
-        """Insert a model item. Let this item's container do it if possible; otherwise do it directly.
+    class TransactionContext:
+        def __init__(self, profile: Profile) -> None:
+            self.__profile = profile
 
-        Passing responsibility to this item's container allows the library to easily track dependencies.
-        However, if this item isn't yet in the library hierarchy, then do the operation directly.
-        """
-        if self.container:
-            self.container.insert_model_item(container, name, before_index, item)
-        else:
-            container.insert_item(name, before_index, item)
+        def __enter__(self) -> Profile.TransactionContext:
+            profile_storage_system = self.__profile._profile_storage_system
+            if profile_storage_system:
+                profile_storage_system.enter_write_delay(self.__profile)
+            for project in self.__profile.projects:
+                project.project_storage_system.enter_transaction()
+            return self
 
-    def remove_model_item(self, container, name, item, *, safe: bool = False) -> Changes.UndeleteLog:
-        """Remove a model item. Let this item's container do it if possible; otherwise do it directly.
+        def __exit__(self, exception_type: typing.Optional[typing.Type[BaseException]],
+                     value: typing.Optional[BaseException],
+                     traceback: typing.Optional[types.TracebackType]) -> typing.Optional[bool]:
+            profile_storage_system = self.__profile._profile_storage_system
+            if profile_storage_system:
+                profile_storage_system.exit_write_delay(self.__profile)
+                profile_storage_system.rewrite_item(self.__profile)
+            for project in self.__profile.projects:
+                project.project_storage_system.exit_transaction()
+            return None
 
-        Passing responsibility to this item's container allows the library to easily track dependencies.
-        However, if this item isn't yet in the library hierarchy, then do the operation directly.
-        """
-        if self.container:
-            return self.container.remove_model_item(container, name, item, safe=safe)
-        else:
-            container.remove_item(name, item)
-            return Changes.UndeleteLog()
-
-    def transaction_context(self):
+    def transaction_context(self) -> contextlib.AbstractContextManager[Profile.TransactionContext]:
         """Return a context object for a document-wide transaction."""
-
-        class Transaction:
-            def __init__(self, profile):
-                self.__profile = profile
-
-            def __enter__(self):
-                self.__profile._profile_storage_system.enter_write_delay(self.__profile)
-                for project in self.__profile.projects:
-                    project.project_storage_system.enter_transaction()
-                return self
-
-            def __exit__(self, type, value, traceback):
-                self.__profile._profile_storage_system.exit_write_delay(self.__profile)
-                self.__profile._profile_storage_system.rewrite_item(self.__profile)
-                for project in self.__profile.projects:
-                    project.project_storage_system.exit_transaction()
-
-        return Transaction(self)
+        return Profile.TransactionContext(self)
 
     def read_profile(self) -> None:
         # read the properties from the storage system. called after open.
@@ -527,7 +592,7 @@ class Profile(Observable.Observable, Persistence.PersistentObject):
             self.storage_system.set_property(self, "version", FileStorageSystem.PROFILE_VERSION)
 
     def read_project(self, project_reference: ProjectReference) -> None:
-        project_reference.load_project(self.profile_context)
+        project_reference.load_project(self.profile_context, self.__cache_dir_path, cache_factory=self.__cache_factory)
 
     def create_project(self, project_dir: pathlib.Path, library_name: str) -> typing.Optional[ProjectReference]:
         project_name = pathlib.Path(library_name)
@@ -579,7 +644,7 @@ class Profile(Observable.Observable, Persistence.PersistentObject):
         else:
             project_reference.close()
             if load:
-                existing_project_reference.load_project(self.profile_context)
+                existing_project_reference.load_project(self.profile_context, self.__cache_dir_path, cache_factory=self.__cache_factory)
             return existing_project_reference
 
     def add_project_index(self, project_path: pathlib.Path, load: bool = True) -> ProjectReference:
@@ -596,9 +661,10 @@ class Profile(Observable.Observable, Persistence.PersistentObject):
         project_reference.establish_last_used()
         return self.add_project_reference(project_reference, load)
 
-    def add_project_memory(self, _uuid: uuid.UUID = None, load: bool = True) -> ProjectReference:
+    def add_project_memory(self, _uuid: typing.Optional[uuid.UUID] = None, load: bool = True) -> ProjectReference:
         assert callable(project_reference_factory_hook)
         project_reference = project_reference_factory_hook("project_memory")
+        assert project_reference
         project_reference.project_uuid = _uuid or uuid.uuid4()
         project_reference.establish_last_used()
         return self.add_project_reference(project_reference, load)
