@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import abc
+import collections
 import copy
 import datetime
 import json
@@ -8,10 +9,17 @@ import os
 import pathlib
 import typing
 import uuid
+import weakref
 
 from nion.utils import Converter
+from nion.utils import DateTime
 from nion.utils import Observable
 
+DictValue = typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any], typing.Tuple[typing.Any], str, float, int, bool, None]
+ItemProxyEntity = typing.Optional[typing.Any]  # use this to make it easy to switch to Entity later
+PersistentDictType = typing.Dict[str, DictValue]
+PersistentMappingType = typing.Mapping[str, DictValue]
+_WeakReferenceType = typing.Any  # Python 3.9+ fix these
 
 ANY = "any"
 STRING = "string"
@@ -31,65 +39,78 @@ REQUIRED = False
 entity_types: typing.Dict[str, EntityType] = dict()
 
 
+def utcnow() -> datetime.datetime:
+    return DateTime.utcnow()
+
+
 def register_entity_type(entity_id: str, entity: EntityType) -> None:
     entity_types[entity_id] = entity
 
-def get_entity_type(entity_id: str) -> EntityType:
-    return entity_types[entity_id]
 
-def build_value(type: str, value) -> typing.Any:
+def unregister_entity_type(entity_id: str) -> None:
+    del entity_types[entity_id]
+
+
+def get_entity_type(entity_id: str) -> typing.Optional[EntityType]:
+    return entity_types.get(entity_id)
+
+
+def build_value(type: str, value: typing.Any) -> typing.Any:
     if value is None:
         return None
-    if type in (SET, ):
+    if type in (SET,):
         return set(value)
-    if type in (BOOLEAN, ):
+    if type in (BOOLEAN,):
         return bool(value)
-    if type in (INT, ):
+    if type in (INT,):
         return int(value)
-    if type in (FLOAT, ):
+    if type in (FLOAT,):
         return float(value)
-    if type in (TIMESTAMP):
+    if type in (TIMESTAMP,):
         return Converter.DatetimeToStringConverter().convert_back(value)
-    if type in (UUID):
+    if type in (UUID,):
         return Converter.UuidToStringConverter().convert_back(value)
-    if type in (PATH):
+    if type in (PATH,):
         return Converter.PathToStringConverter().convert_back(value)
     return value
 
 
-def dict_value(type: str, value):
+def dict_value(type: str, value: typing.Any) -> DictValue:
     if value is None:
         return None
-    if type in (SET, ):
+    if type in (SET,):
         return list(value)
-    if type in (BOOLEAN, ):
+    if type in (BOOLEAN,):
         return bool(value)
-    if type in (INT, ):
+    if type in (INT,):
         return int(value)
-    if type in (FLOAT, ):
+    if type in (FLOAT,):
         return float(value)
-    if type in (TIMESTAMP):
+    if type in (TIMESTAMP,):
         return Converter.DatetimeToStringConverter().convert(value)
-    if type in (UUID):
+    if type in (UUID,):
         return Converter.UuidToStringConverter().convert(value)
-    if type in (PATH):
+    if type in (PATH,):
         return Converter.PathToStringConverter().convert(value)
-    return value
-
-
-DictValue = typing.Optional[typing.Union[typing.Dict, typing.List, typing.Tuple, str, int]]
-
-
-ItemProxyEntity = typing.Optional[typing.Any]  # use this to make it easy to switch to Entity later
+    return typing.cast(DictValue, value)
 
 
 class ItemProxy:
-    def __init__(self):
+    def __init__(self, item_uuid: typing.Optional[uuid.UUID]) -> None:
+        self.__item_uuid = item_uuid
         self.__item = None
 
     def close(self) -> None:
         pass
 
+    @property
+    def item_uuid(self) -> typing.Optional[uuid.UUID]:
+        return self.__item_uuid
+
+    @item_uuid.setter
+    def item_uuid(self, value: typing.Optional[uuid.UUID]) -> None:
+        self.__item_uuid = value
+
     @property
     def item(self) -> ItemProxyEntity:
         return self.__item
@@ -101,22 +122,69 @@ class ItemProxy:
 
 class EntityContext(abc.ABC):
 
+    @abc.abstractmethod
+    def register(self, entity: Entity) -> None: ...
+
+    @abc.abstractmethod
+    def unregister(self, entity: Entity) -> None: ...
+
     @abc.abstractmethod
     def create_item_reference(self, item_uuid: typing.Optional[uuid.UUID] = None) -> ItemProxy: ...
 
+    @abc.abstractmethod
+    def update_item_reference_uuid(self, item_reference: ItemProxy, item_uuid: typing.Optional[uuid.UUID]) -> None: ...
+
 
 class SimpleEntityContext(EntityContext):
+    def __init__(self) -> None:
+        self.__entity_map: typing.Dict[uuid.UUID, _WeakReferenceType] = dict()
+        self.__item_proxies: typing.List[_WeakReferenceType] = list()
+
+    def register(self, entity: Entity) -> None:
+        self.__entity_map[entity.uuid] = weakref.ref(entity)
+        for item_proxy_ref in self.__item_proxies:
+            item_proxy = item_proxy_ref()
+            if item_proxy and item_proxy.item_uuid == entity.uuid:
+                item_proxy.item = entity
+
+    def unregister(self, entity: Entity) -> None:
+        self.__entity_map.pop(entity.uuid)
+        for item_proxy_ref in self.__item_proxies:
+            item_proxy = item_proxy_ref()
+            if item_proxy and item_proxy.item_uuid == entity.uuid:
+                item_proxy.item = None
 
     def create_item_reference(self, item_uuid: typing.Optional[uuid.UUID] = None) -> ItemProxy:
-        return ItemProxy()
+        item_proxy = ItemProxy(item_uuid)
+        if item_uuid and item_uuid in self.__entity_map:
+            entity_ref = self.__entity_map.get(item_uuid)
+            if entity_ref:
+                item_proxy.item = entity_ref()
+        item_proxy_ref = weakref.ref(item_proxy)
+        self.__item_proxies.append(item_proxy_ref)
+
+        def finalize(item_proxy_ref: _WeakReferenceType, item_proxies: typing.List[_WeakReferenceType]) -> None:
+            item_proxies.pop(item_proxies.index(item_proxy_ref))
+
+        weakref.finalize(item_proxy, finalize, item_proxy_ref, self.__item_proxies)
+        return item_proxy
+
+    def update_item_reference_uuid(self, item_proxy: ItemProxy, item_uuid: typing.Optional[uuid.UUID]) -> None:
+        if item_uuid and item_uuid in self.__entity_map:
+            entity_ref = self.__entity_map.get(item_uuid)
+            if entity_ref:
+                item_proxy.item = entity_ref()
+                return
+        item_proxy.item = None
+
 
 
 class Field(abc.ABC):
-    """A field in an entity or another field.
+    """A value-holding field in an entity or another field.
 
     The context is used to resolve references and must be valid to read entities containing references.
     """
-    def __init__(self, context: typing.Optional[EntityContext]):
+    def __init__(self, context: typing.Optional[EntityContext]) -> None:
         self.__context = context
 
     def close(self) -> None:
@@ -133,7 +201,7 @@ class Field(abc.ABC):
 
         Subclasses should propagate the context to their sub fields.
         """
-        assert (self.__context is None) != (context is None)  # one or the other is None
+        assert (self.__context is None) != (context is None), f"{self.__context} {context}"
         self.__context = context
 
     @abc.abstractmethod
@@ -151,9 +219,17 @@ class Field(abc.ABC):
     def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
         pass
 
+    def field_by_key(self, key: str) -> Field:
+        raise AttributeError()
+
+    def field_by_index(self, index: int) -> Field:
+        raise AttributeError()
+
 
 class PropertyField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], type: str, optional: bool, default):
+    """A value property field, which can hold a simple value."""
+
+    def __init__(self, context: typing.Optional[EntityContext], type: str, optional: bool, default: typing.Any) -> None:
         super().__init__(context)
         self.__type = type
         self.__optional = optional
@@ -176,7 +252,9 @@ class PropertyField(Field):
 
 
 class TupleField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], type: FieldType, optional: bool, default_values):
+    """A tuple property field, which can hold an indefinite tuple of simple values."""
+
+    def __init__(self, context: typing.Optional[EntityContext], type: FieldType, optional: bool, default_values: typing.Optional[typing.Sequence[typing.Any]]) -> None:
         super().__init__(context)
         assert isinstance(type, FieldType)
         assert default_values is None or isinstance(default_values, (tuple, list))
@@ -190,7 +268,7 @@ class TupleField(Field):
         self.__clear()
         super().close()
 
-    def __clear(self):
+    def __clear(self) -> None:
         if self.__fields is not None:
             for field in self.__fields:
                 field.close()
@@ -224,7 +302,9 @@ class TupleField(Field):
 
 
 class FixedTupleField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], types: typing.Sequence[FieldType], optional: bool, default_values):
+    """A tuple property field, which can hold a fixed length tuple of simple values."""
+
+    def __init__(self, context: typing.Optional[EntityContext], types: typing.Sequence[FieldType], optional: bool, default_values: typing.Optional[typing.Tuple[typing.Any, ...]]) -> None:
         super().__init__(context)
         assert default_values is None or isinstance(default_values, (tuple, list))
         self.__types = types
@@ -237,7 +317,7 @@ class FixedTupleField(Field):
         self.__clear()
         super().close()
 
-    def __clear(self):
+    def __clear(self) -> None:
         if self.__fields is not None:
             for field in self.__fields:
                 field.close()
@@ -272,7 +352,9 @@ class FixedTupleField(Field):
 
 
 class RecordField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], field_type_map: typing.Mapping[str, FieldType]):
+    """A record property field, which can hold a mapping of strings to simple values."""
+
+    def __init__(self, context: typing.Optional[EntityContext], field_type_map: typing.Mapping[str, FieldType]) -> None:
         super().__init__(context)
         self.__field_type_map = field_type_map
         self.__field_map: typing.Dict[str, Field] = dict()
@@ -306,14 +388,21 @@ class RecordField(Field):
 
     @property
     def field_value(self) -> typing.Any:
-        return None
+        d = {k: field.field_value for k, field in self.__field_map.items()}
+        return collections.namedtuple("record", d.keys())(*d.values())
 
     def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
-        pass
+        assert isinstance(value, dict)
+        for field in self.__field_map.values():
+            field.close()
+        for k, field_value in value.items():
+            self.__field_map[k] = self.__field_type_map[k].create_and_read(self._context, field_value)
 
 
 class ArrayField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], type: FieldType, optional: bool):
+    """A array of fields, which can be simple values, references, or components."""
+
+    def __init__(self, context: typing.Optional[EntityContext], type: FieldType, optional: bool) -> None:
         super().__init__(context)
         self.__type = type
         self.__optional = optional
@@ -345,7 +434,7 @@ class ArrayField(Field):
         return l if l or not self.__optional else None
 
     @property
-    def field_value(self) -> typing.Any:
+    def field_value(self) -> typing.Sequence[typing.Any]:
         return [field.field_value for field in self.__fields]
 
     def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
@@ -354,7 +443,6 @@ class ArrayField(Field):
 
     def insert_value(self, container: ItemProxyEntity, index: int, value: typing.Any) -> None:
         if isinstance(value, Entity):
-            assert not value._container
             field = self.__type.create(self._context)
             field.set_field_value(container, value)  # no container yet
             self.__fields.insert(index, field)
@@ -365,9 +453,14 @@ class ArrayField(Field):
         field = self.__fields.pop(index)
         field.set_field_value(None, None)  # no container
 
+    def field_by_index(self, index: int) -> Field:
+        return self.__fields[index]
+
 
 class MapField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], key: FieldType, value: FieldType, optional: bool):
+    """A map of strings to fields, which can be simple values, references, or components."""
+
+    def __init__(self, context: typing.Optional[EntityContext], key: FieldType, value: FieldType, optional: bool) -> None:
         super().__init__(context)
         self.__key = key
         self.__value = value
@@ -407,16 +500,27 @@ class MapField(Field):
         assert isinstance(value, dict)
         self.__fields = {k: self.__value.create_and_read(self._context, v) for k, v in value.items()}
 
+    def set_value(self, container: ItemProxyEntity, key: str, value: typing.Any) -> None:
+        if isinstance(value, Entity):
+            field = self.__value.create(self._context)
+            field.set_field_value(container, value)  # no container yet
+            self.__map[key] = field
+        else:
+            raise IndexError()
+
 
 class ReferenceField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], type: EntityType):
+    """A reference field, references another entity without cascading delete."""
+
+    def __init__(self, context: typing.Optional[EntityContext], type: EntityType) -> None:
         super().__init__(context)
         self.__type = type
-        self.__reference_uuid: typing.Optional[str] = None
+        self.__reference_uuid: typing.Optional[uuid.UUID] = None
         self.__reference: typing.Optional[ItemProxy] = None  # proxy is only valid when context is valid
         self.__shadow_item: typing.Optional[Entity] = None  # used when proxy is None
-        if self._context:
-            self.__reference = self._context.create_item_reference()
+        if context:
+            super().set_context(None)  # unset it safely
+            self.set_context(context)
 
     def close(self) -> None:
         if self.__reference:
@@ -428,7 +532,7 @@ class ReferenceField(Field):
         super().set_context(context)
         if self._context:
             if not self.__reference:
-                self.__reference = self._context.create_item_reference(uuid.UUID(self.__reference_uuid) if self.__reference_uuid else None)
+                self.__reference = self._context.create_item_reference(self.__reference_uuid if self.__reference_uuid else None)
             if self.__shadow_item:
                 self.__reference.item = self.__shadow_item
                 self.__shadow_item = None
@@ -438,11 +542,13 @@ class ReferenceField(Field):
             self.__reference = None
 
     def read(self, dict_value: typing.Any) -> Field:
-        self.__reference_uuid = copy.deepcopy(dict_value)
+        self.__reference_uuid = uuid.UUID(copy.deepcopy(dict_value))
+        if self.__reference and self._context:
+            self._context.update_item_reference_uuid(self.__reference, self.__reference_uuid)
         return self
 
     def write(self) -> DictValue:
-        return copy.deepcopy(self.__reference_uuid)
+        return str(self.__reference_uuid) if self.__reference_uuid else None
 
     @property
     def field_value(self) -> typing.Any:
@@ -455,7 +561,7 @@ class ReferenceField(Field):
         item = typing.cast(Entity, value)
         if item:
             item_uuid = item.uuid  # prefer _get_field_value; but use direct accessor for legacy Swift PersistentObject compatibility
-            self.__reference_uuid = str(item_uuid)
+            self.__reference_uuid = item_uuid
             if self.__reference:
                 self.__reference.item = item
             else:
@@ -469,11 +575,12 @@ class ReferenceField(Field):
 
 
 class ComponentField(Field):
-    def __init__(self, context: typing.Optional[EntityContext], entity_id: str):
+    """A component field, references another entity with cascading delete."""
+
+    def __init__(self, context: typing.Optional[EntityContext], entity_id: str) -> None:
         super().__init__(context)
         self.__type = get_entity_type(entity_id)
         self.__entity: typing.Optional[Entity] = None
-        self.__value: ItemProxyEntity = None
 
     def close(self) -> None:
         if self.__entity:
@@ -487,8 +594,8 @@ class ComponentField(Field):
             self.__entity._set_entity_context(context)
 
     def read(self, dict_value: typing.Any) -> Field:
-        self.__entity = self.__type.create(self._context)
-        self.__entity.read(dict_value)
+        assert self.__type
+        self.__entity = self.__type.create(self._context, dict_value)
         return self
 
     def write(self) -> DictValue:
@@ -496,20 +603,46 @@ class ComponentField(Field):
 
     @property
     def field_value(self) -> typing.Any:
-        return self.__value
+        return self.__entity
 
     def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
-        if self.__value:
-            self.__value._container = None
-            self.__value._set_entity_context(None)
-        self.__value = value
-        if self.__value:
-            self.__value._container = container
-            self.__value._set_entity_context(self._context)
+        assert value is None or not value._container
+        if self.__entity:
+            self.__entity._container = None
+            if self.__entity._entity_context:
+                self.__entity._set_entity_context(None)
+        self.__entity = value
+        if self.__entity:
+            self.__entity._container = container
+            if self._context:
+                self.__entity._set_entity_context(self._context)
+
+    def field_by_key(self, key: str) -> Field:
+        assert self.__entity
+        return self.__entity.get_field(key)
+
+
+class ComponentPlaceholderField(Field):
+    """Keep the dict, but otherwise be non-functional."""
+
+    def __init__(self, context: typing.Optional[EntityContext]) -> None:
+        super().__init__(context)
+        self.__d = typing.cast(DictValue, None)
+
+    def read(self, dict_value: typing.Any) -> Field:
+        self.__d = copy.deepcopy(dict_value)
+        return self
+
+    def write(self) -> DictValue:
+        return copy.deepcopy(self.__d)
 
 
 class FieldType(abc.ABC):
-    def __init__(self, field_class: typing.Callable[..., Field], *args, **kwargs):
+    """A field type that provides the ability to create a field of a given type.
+
+    The field class is a callable returning a field. The args and kwargs are passed to the callable.
+    """
+    def __init__(self, field_class: typing.Callable[..., Field], *args: typing.Any, **kwargs: typing.Any) -> None:
         assert callable(field_class)
         self.__field_class = field_class
         self.__args = args
@@ -520,14 +653,14 @@ class FieldType(abc.ABC):
         return self.__field_class
 
     @property
-    def _args(self):
+    def _args(self) -> typing.Tuple[typing.Any, ...]:
         return self.__args
 
     @property
-    def _kwargs(self):
+    def _kwargs(self) -> typing.Dict[str, typing.Any]:
         return self.__kwargs
 
-    def _call(self, context: typing.Optional[EntityContext], field_class: typing.Callable[..., Field], *args, **kwargs) -> Field:
+    def _call(self, context: typing.Optional[EntityContext], field_class: typing.Callable[..., Field], *args: typing.Any, **kwargs: typing.Any) -> Field:
         return field_class(context, *args, **kwargs)
 
     def create(self, context: typing.Optional[EntityContext]) -> Field:
@@ -538,7 +671,7 @@ class FieldType(abc.ABC):
 
 
 class PropertyType(FieldType):
-    def __init__(self, type: str, optional: bool, default):
+    def __init__(self, type: str, optional: bool, default: typing.Any) -> None:
         super().__init__(PropertyField, type, optional, default)
         self.type = type
         self.optional = optional
@@ -546,34 +679,34 @@ class PropertyType(FieldType):
 
 
 class TupleType(FieldType):
-    def __init__(self, type: FieldType, optional: bool, default):
+    def __init__(self, type: FieldType, optional: bool, default: typing.Optional[typing.Sequence[typing.Any]]) -> None:
         super().__init__(TupleField, type, optional, default)
         self.type = type
         self.optional = optional
 
 
 class FixedTupleType(FieldType):
-    def __init__(self, types: typing.Sequence[FieldType], optional: bool, default):
+    def __init__(self, types: typing.Sequence[FieldType], optional: bool, default: typing.Optional[typing.Tuple[typing.Any, ...]]) -> None:
         super().__init__(FixedTupleField, types, optional, default)
         self.types = list(types)
         self.optional = optional
 
 
 class RecordType(FieldType):
-    def __init__(self, field_type_map: typing.Mapping[str, FieldType]):
+    def __init__(self, field_type_map: typing.Mapping[str, FieldType]) -> None:
         super().__init__(RecordField, field_type_map)
         self.__field_type_map = dict(field_type_map)
 
 
 class ArrayType(FieldType):
-    def __init__(self, type: FieldType, optional: bool):
+    def __init__(self, type: FieldType, optional: bool) -> None:
         super().__init__(ArrayField, type, optional)
         self.type = type
         self.optional = optional
 
 
 class MapType(FieldType):
-    def __init__(self, key: str, value: FieldType, optional: bool):
+    def __init__(self, key: str, value: FieldType, optional: bool) -> None:
         super().__init__(MapField, key, value, optional)
         self.key = key
         self.value = value
@@ -581,81 +714,109 @@ class MapType(FieldType):
 
 
 class ReferenceType(FieldType):
-    def __init__(self, type: typing.Optional[EntityType]):
+    def __init__(self, type: typing.Optional[EntityType]) -> None:
         super().__init__(ReferenceField, type)
         self.type = type
 
 
 class ComponentType(FieldType):
-    def __init__(self, entity_id: str):
+    def __init__(self, entity_id: str, required: bool) -> None:
         super().__init__(ComponentField, entity_id)
         self.entity_id = entity_id
+        self.required = required
 
     def create_and_read(self, context: typing.Optional[EntityContext], dict_value: DictValue) -> Field:
         assert isinstance(dict_value, dict)
         d = dict_value
         entity_type = get_entity_type(d["type"])
-        assert entity_type
-        field_type = entity_type.entity_id if entity_type else self._args[0]
-        return self._call(context, self._field_class, *((field_type, ) + self._args[1:]), **self._kwargs).read(dict_value)
+        if entity_type:
+            field_type = entity_type.entity_id if entity_type else self._args[0]
+            return self._call(context, self._field_class, *((field_type, ) + self._args[1:]), **self._kwargs).read(dict_value)
+        else:
+            return ComponentPlaceholderField(context).read(dict_value)
 
 
-EntityTransforms = typing.Tuple[typing.Callable[[typing.Dict], typing.Dict], typing.Callable[[typing.Dict], typing.Dict]]
+_EntityTransform = typing.Callable[[PersistentDictType], PersistentDictType]
+_EntityTransforms = typing.Tuple[_EntityTransform, _EntityTransform]
 
 
 class Entity(Observable.Observable):
     """An instance of an entity type.
 
+    To support deepcopy, subclasses must be able to be initialized with no arguments (passing entity_type to this init)
+    or override deepcopy.
     """
-    def __init__(self, *,
-                 type: EntityType,
-                 version: typing.Optional[int],
-                 context: typing.Optional[EntityContext],
-                 field_type_map: typing.Mapping[str, FieldType],
-                 renames: typing.Mapping[str, str],
-                 transforms: EntityTransforms):
+    def __init__(self, entity_type: typing.Optional[EntityType] = None, context: typing.Optional[EntityContext] = None) -> None:
         super().__init__()
-        self.__context = context
-        self.__entity_type = type
-        self.__version = version
-        self.__field_type_map = field_type_map
+        assert entity_type, str(self.__class__)
+        self.__context: typing.Optional[EntityContext] = None
+        self.__entity_type = entity_type
+        self.__version = entity_type._version
+        self.__field_type_map = entity_type._field_type_map
         self.__field_dict : typing.Dict[str, Field] = dict()
-        self.__renames = renames
-        self.__transforms = transforms
+        self.__renames = entity_type._renames
+        self.__transforms = entity_type._transforms
         self._container = None
         for field_name, field_type in self.__field_type_map.items():
             self.__field_dict[field_name] = field_type.create(self.__context)
         self._set_field_value("uuid", uuid.uuid4())
-        self._set_field_value("modified", datetime.datetime.utcnow())
+        self._set_field_value("modified", utcnow())
+        if context:
+            self._set_entity_context(context)
 
     def close(self) -> None:
         pass
 
-    def __deepcopy__(self, memo):
-        entity_copy = self.__class__()
-        for k, field in self.__field_dict.items():
-            entity_copy._set_field_value(k, self._get_field_value(k))
+    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> Entity:
+        entity_copy = self._deepcopy()
         memo[id(self)] = entity_copy
         return entity_copy
 
+    def __repr__(self) -> str:
+        return f"entity '{self.__entity_type.entity_id}' at 0x{id(self):x}"
+
+    def _create(self, context: typing.Optional[EntityContext]) -> Entity:
+        return self.__class__(self.__entity_type, context)
+
+    def _deepcopy(self) -> Entity:
+        context = SimpleEntityContext()
+        entity_copy = self._create(context)
+        entity_copy.read(self.write_to_dict())
+        entity_copy._set_entity_context(None)
+        return entity_copy
+
     @property
     def _entity_context(self) -> typing.Optional[EntityContext]:
         return self.__context
 
     def _set_entity_context(self, context: typing.Optional[EntityContext]) -> None:
-        assert (self.__context is None) != (context is None)  # one or the other is None
+        assert (self.__context is None) != (context is None), f"{self.__context} {context}"
+        if self.__context:
+            self.__context.unregister(self)
         self.__context = context
+        if self.__context:
+            self.__context.register(self)
         for field in self.__field_dict.values():
             field.set_context(context)
 
-    def read(self, properties: typing.Mapping) -> Entity:
+    def read(self, properties: PersistentMappingType) -> Entity:
+        # unregister old uuid
+        if self.__context:
+            self.__context.unregister(self)
         properties = self.__transforms[0](dict(properties))  # transform forward
         for field_name, field_type in self.__field_type_map.items():
             d = properties.get(self.__renames.get(field_name, field_name))
-            self.__field_dict[field_name].read(d)
+            if d is not None:
+                self.__field_dict[field_name].read(d)
+        # register new uuid
+        if self.__context:
+            self.__context.register(self)
         return self
 
-    def write_to_dict(self) -> typing.Dict:
+    def write(self) -> DictValue:
+        return self.write_to_dict()
+
+    def write_to_dict(self) -> PersistentDictType:
         d: typing.Dict[str, DictValue] = dict()
         d["type"] = self.__entity_type.entity_id
         if self.__version:
@@ -692,6 +853,9 @@ class Entity(Observable.Observable):
     def __get_field(self, name: str) -> typing.Optional[Field]:
         return self.__field_dict.get(name)
 
+    def get_field(self, name: str) -> Field:
+        return typing.cast(Field, self.__get_field(name))
+
     def _get_field_value(self, name: str) -> typing.Any:
         field = self.__get_field(name)
         if field:
@@ -703,7 +867,8 @@ class Entity(Observable.Observable):
         if field:
             field.set_field_value(self, value)
             self._field_value_changed(name, field.write())
-            self._set_modified(datetime.datetime.utcnow())
+            if name not in ("modified", "uuid"):
+                self._set_modified(utcnow())
             self.property_changed_event.fire(name)
         else:
             raise AttributeError()
@@ -739,7 +904,7 @@ class Entity(Observable.Observable):
     def _remove_item(self, name: str, item: ItemProxyEntity) -> None:
         array_field = typing.cast(typing.Optional[ArrayField], self.__get_field(name))
         if array_field:
-            index = typing.cast(typing.List, self._get_array_items(name)).index(item)
+            index = list(self._get_array_items(name)).index(item)
             array_field.remove_value_at_index(index)  # passing self for container
             self.item_removed_event.fire(name, item, index)
         else:
@@ -750,7 +915,7 @@ class Entity(Observable.Observable):
     def begin_reading(self) -> None:
         pass
 
-    def read_from_dict(self, properties: typing.Dict) -> None:
+    def read_from_dict(self, properties: PersistentMappingType) -> None:
         self.read(properties)
 
     def finish_reading(self) -> None:
@@ -763,40 +928,80 @@ class Entity(Observable.Observable):
         self._container = None
 
     @property
-    def container(self):
+    def container(self) -> typing.Any:
         return self._container
 
     @property
-    def persistent_object_context(self):
+    def persistent_object_context(self) -> typing.Any:
         return self._entity_context
 
     @persistent_object_context.setter
-    def persistent_object_context(self, value):
+    def persistent_object_context(self, value: typing.Any) -> None:
         self._set_entity_context(value)
 
     @property
-    def item_names(self) -> typing.List:
-        return list()
+    def item_names(self) -> typing.List[str]:
+        return [k for k, f in self.__field_type_map.items() if isinstance(f, ComponentType)]
 
     @property
-    def relationship_names(self) -> typing.List:
-        return list()
+    def relationship_names(self) -> typing.List[str]:
+        return [k for k, f in self.__field_type_map.items() if isinstance(f, ArrayType)]
+
+    def item_specifier(self) -> typing.Any:
+        class ItemSpecifier:
+
+            def __init__(self, *, item: typing.Optional[Entity] = None, item_uuid: typing.Optional[uuid.UUID] = None) -> None:
+                self.__item_uuid = typing.cast(uuid.UUID, item.uuid) if item else item_uuid
+                assert (self.__item_uuid is None) or isinstance(self.__item_uuid, uuid.UUID)
+
+            def __hash__(self) -> typing.Any:
+                return hash(self.__item_uuid)
+
+            def __eq__(self, other: typing.Any) -> bool:
+                if isinstance(other, self.__class__):
+                    return bool(self.__item_uuid == other.__item_uuid)
+                return False
+
+            @property
+            def item_uuid(self) -> typing.Optional[uuid.UUID]:
+                return self.__item_uuid
+
+            def write(self) -> typing.Optional[DictValue]:
+                if self.__item_uuid:
+                    return str(self.__item_uuid)
+                return None
+
+            @staticmethod
+            def read(d: DictValue) -> typing.Optional[ItemSpecifier]:
+                if isinstance(d, str):
+                    return ItemSpecifier(item_uuid=uuid.UUID(d))
+                elif isinstance(d, uuid.UUID):
+                    return ItemSpecifier(item_uuid=d)
+                elif isinstance(d, dict) and "item_uuid" in d:
+                    return ItemSpecifier(item_uuid=uuid.UUID(d["item_uuid"]))
+                elif isinstance(d, dict) and "uuid" in d:
+                    return ItemSpecifier(item_uuid=uuid.UUID(d["uuid"]))
+                else:
+                    return None
+
+        return ItemSpecifier(item=self)
 
 
-def no_transform(x: typing.Dict) -> typing.Dict:
+def no_transform(x: PersistentDictType) -> PersistentDictType:
     return x
 
 
 class EntityType:
-    def __init__(self, entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Mapping[str, FieldType]):
+    def __init__(self, entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Mapping[str, FieldType], factory: typing.Optional[typing.Callable[[EntityType, typing.Optional[EntityContext]], Entity]] = None) -> None:
         self.__entity_id = entity_id
         self.__base = base
         self.__version = version
         self.__renames: typing.Dict[str, str] = dict()
-        self.__transforms: EntityTransforms = (no_transform, no_transform)
+        self.__transforms: _EntityTransforms = (no_transform, no_transform)
         self.__field_type_map: typing.Dict[str, FieldType] = dict()
         self.__field_type_map["uuid"] = prop(UUID)
         self.__field_type_map["modified"] = prop(TIMESTAMP)
+        self.__factory = factory
         while base:
             for k, v in base._field_type_map.items():
                 if not k in ("uuid", "modified"):
@@ -808,6 +1013,10 @@ class EntityType:
             self.__field_type_map[k] = v
         register_entity_type(entity_id, self)
 
+    @property
+    def _version(self) -> typing.Optional[int]:
+        return self.__version
+
     @property
     def _field_type_map(self) -> typing.Mapping[str, FieldType]:
         return self.__field_type_map
@@ -816,22 +1025,23 @@ class EntityType:
     def _renames(self) -> typing.Mapping[str, str]:
         return self.__renames
 
-    def create(self, context: typing.Optional[EntityContext] = None, d: typing.Optional[typing.Dict] = None) -> Entity:
-        entity = Entity(**self.entity_init_kwargs(context))
+    @property
+    def _transforms(self) -> _EntityTransforms:
+        return self.__transforms
+
+    def create(self, context: typing.Optional[EntityContext] = None, d: typing.Optional[PersistentMappingType] = None) -> Entity:
+        type = d.get("type", None) if d else None
+        if type:
+            concrete_subclasses = self.concrete_subclasses
+            if concrete_subclasses:
+                for concrete_subclass in concrete_subclasses:
+                    if type == concrete_subclass.entity_id:
+                        return concrete_subclass.create(context, d)
+        entity = self.__factory(self, context) if callable(self.__factory) else Entity(self, context)
         if d is not None:
             entity.read(d)
         return entity
 
-    def entity_init_kwargs(self, context: typing.Optional[EntityContext] = None) -> typing.Dict[str, typing.Any]:
-        return {
-            "type": self,
-            "version": self.__version,
-            "context": context,
-            "field_type_map": self.__field_type_map,
-            "renames": self.__renames,
-            "transforms": self.__transforms,
-        }
-
     @property
     def base(self) -> typing.Optional[EntityType]:
         return self.__base
@@ -843,7 +1053,7 @@ class EntityType:
     def rename(self, field_name: str, storage_field_name: str) -> None:
         self.__renames[field_name] = storage_field_name
 
-    def transform(self, forward: typing.Callable[[typing.Dict], typing.Dict], backward: typing.Callable[[typing.Dict], typing.Dict]) -> None:
+    def transform(self, forward: _EntityTransform, backward: _EntityTransform) -> None:
         self.__transforms = (forward, backward)
 
     def is_subclass_of(self, entity_type: EntityType) -> bool:
@@ -858,13 +1068,13 @@ class EntityType:
         return [entity_type for entity_type in entity_types.values() if entity_type.is_subclass_of(self) and not entity_type.subclasses]
 
 
-def prop(type: str, optional: bool=False, *, default=None) -> PropertyType:
+def prop(type: str, optional: bool = False, *, default: typing.Any = None) -> PropertyType:
     return PropertyType(type, optional, default)
 
-def indefinite_tuple(type: FieldType, optional: bool=False, default=None) -> TupleType:
+def indefinite_tuple(type: FieldType, optional: bool = False, default: typing.Optional[typing.Sequence[typing.Any]] = None) -> TupleType:
     return TupleType(type, optional, default)
 
-def fixed_tuple(types: typing.List[FieldType], optional: bool = False, default=None) -> FixedTupleType:
+def fixed_tuple(types: typing.List[FieldType], optional: bool = False, default: typing.Optional[typing.Tuple[typing.Any, ...]] = None) -> FixedTupleType:
     return FixedTupleType(types, optional, default)
 
 def record(field_type_map: typing.Dict[str, FieldType]) -> RecordType:
@@ -879,16 +1089,16 @@ def map(key: str, value: FieldType, optional: bool = False) -> MapType:
 def reference(type: typing.Optional[EntityType] = None) -> ReferenceType:
     return ReferenceType(type)
 
-def component(type: typing.Union[EntityType, str]) -> ComponentType:
+def component(type: typing.Union[EntityType, str], required: bool = True) -> ComponentType:
     if isinstance(type, EntityType):
-        return ComponentType(type.entity_id)
+        return ComponentType(type.entity_id, required)
     else:  # str, used for forward or self references
-        return ComponentType(type)
+        return ComponentType(type, required)
 
-def entity(entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Dict[str, FieldType]) -> EntityType:
-    return EntityType(entity_id, base, version, field_type_map)
+def entity(entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Dict[str, FieldType], factory: typing.Optional[typing.Callable[[EntityType, typing.Optional[EntityContext]], Entity]] = None) -> EntityType:
+    return EntityType(entity_id, base, version, field_type_map, factory)
 
-def read_json(path: pathlib.Path) -> typing.Dict:
+def read_json(path: pathlib.Path) -> PersistentDictType:
     """Read the json file from path. Rename path if not readable."""
     properties = dict()
     if path and path.exists():
