@@ -4,6 +4,7 @@ import datetime
 import json
 import logging
 import os
+import pathlib
 import unittest
 import uuid
 
@@ -17,6 +18,7 @@ from nion.swift.model import DataItem
 from nion.swift.model import ImportExportManager
 from nion.swift.model import Utility
 from nion.swift.test import TestContext
+from nion.utils import DateTime
 
 
 class TestImportExportManagerClass(unittest.TestCase):
@@ -82,10 +84,10 @@ class TestImportExportManagerClass(unittest.TestCase):
             data_item = DataItem.DataItem(numpy.zeros((16, 16), dtype=numpy.double))
             document_model.append_data_item(data_item)
             display_item = document_model.get_display_item_for_data_item(data_item)
-            handler.write_display_item(None, display_item, file_path, "ndata")
+            handler.write_display_item(display_item, pathlib.Path(file_path), "ndata")
             self.assertTrue(os.path.exists(file_path))
             try:
-                data_items = handler.read_data_items(None, "ndata", file_path)
+                data_items = handler.read_data_items("ndata", pathlib.Path(file_path))
                 self.assertEqual(len(data_items), 1)
                 data_item = data_items[0]
                 for data_item in data_items:
@@ -104,7 +106,7 @@ class TestImportExportManagerClass(unittest.TestCase):
                 json.dump({"version": 1}, f)
             handler = ImportExportManager.NumPyImportExportHandler("numpy-io-handler", "npy", ["npy"])
             try:
-                data_items = handler.read_data_items(None, "npy", file_path_npy)
+                data_items = handler.read_data_items("npy", pathlib.Path(file_path_npy))
                 self.assertEqual(len(data_items), 1)
                 data_item = data_items[0]
                 for data_item in data_items:
@@ -113,6 +115,27 @@ class TestImportExportManagerClass(unittest.TestCase):
                 os.remove(file_path_npy)
                 os.remove(file_path_json)
 
+    def test_standard_io_write_to_then_read_from_temp_file(self):
+        with TestContext.create_memory_context() as test_context:
+            document_model = test_context.create_document_model()
+            current_working_directory = os.getcwd()
+            extensions = ("jpeg", "png", "gif", "bmp", "webp")
+            for extension in extensions:
+                file_path = os.path.join(current_working_directory, f"__file.{extension}")
+                handler = ImportExportManager.StandardImportExportHandler(f"{extension}-io-handler", extension, [extension])
+                data_item = DataItem.DataItem(numpy.zeros((16, 16), dtype=numpy.double))
+                document_model.append_data_item(data_item)
+                display_item = document_model.get_display_item_for_data_item(data_item)
+                handler.write_display_item(display_item, pathlib.Path(file_path), extension)
+                self.assertTrue(os.path.exists(file_path))
+                try:
+                    data_items = handler.read_data_items(extension, pathlib.Path(file_path))
+                    self.assertEqual(len(data_items), 1)
+                    for data_item in data_items:
+                        data_item.close()
+                finally:
+                    os.remove(file_path)
+
     def test_get_writers_for_empty_data_item_returns_valid_list(self):
         data_item = DataItem.DataItem()
         with contextlib.closing(data_item):
@@ -133,7 +156,7 @@ class TestImportExportManagerClass(unittest.TestCase):
         with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
             self.assertIsNotNone(data_item.created)
             self.assertEqual(data_item.timezone_offset, "+0300")
-            local_offset_seconds = int(round((datetime.datetime.now() - datetime.datetime.utcnow()).total_seconds()))
+            local_offset_seconds = int(round((datetime.datetime.now() - DateTime.utcnow()).total_seconds()))
             # check both matches for DST
             match1 = datetime.datetime(year=2015, month=6, day=10, hour=19 - 3, minute=31, second=52, microsecond=780511) + datetime.timedelta(seconds=local_offset_seconds)
             match2 = datetime.datetime(year=2015, month=6, day=10, hour=19 - 3, minute=31, second=52, microsecond=780511) + datetime.timedelta(seconds=local_offset_seconds + 3600)
@@ -190,7 +213,7 @@ class TestImportExportManagerClass(unittest.TestCase):
         self.assertNotEqual(id(new_xdata.intensity_calibration), id(intensity_calibration))
         self.assertEqual(new_xdata.intensity_calibration, intensity_calibration)
         self.assertNotEqual(id(new_xdata.dimensional_calibrations[0]), id(dimensional_calibrations[0]))
-        self.assertEqual(new_xdata.dimensional_calibrations, dimensional_calibrations)
+        self.assertEqual(tuple(new_xdata.dimensional_calibrations), tuple(dimensional_calibrations))
         self.assertNotEqual(id(new_xdata.metadata), id(metadata))
         self.assertEqual(new_xdata.metadata, metadata)
         self.assertNotEqual(id(new_xdata.data_descriptor), id(data_descriptor))
@@ -262,7 +285,7 @@ class TestImportExportManagerClass(unittest.TestCase):
             current_working_directory = os.getcwd()
             file_path = os.path.join(current_working_directory, "__file.csv")
             handler = ImportExportManager.CSV1ImportExportHandler("csv1-io-handler", "CSV 1D", ["csv"])
-            handler.write_display_item(None, display_item, file_path, "csv")
+            handler.write_display_item(display_item, pathlib.Path(file_path), "csv")
             self.assertTrue(os.path.exists(file_path))
             try:
                 saved_data = numpy.genfromtxt(file_path, delimiter=", ")
@@ -290,7 +313,7 @@ class TestImportExportManagerClass(unittest.TestCase):
             current_working_directory = os.getcwd()
             file_path = os.path.join(current_working_directory, "__file.csv")
             handler = ImportExportManager.CSV1ImportExportHandler("csv1-io-handler", "CSV 1D", ["csv"])
-            handler.write_display_item(None, display_item, file_path, "csv")
+            handler.write_display_item(display_item, pathlib.Path(file_path), "csv")
             self.assertTrue(os.path.exists(file_path))
             try:
                 saved_data = numpy.genfromtxt(file_path, delimiter=", ")
@@ -342,7 +365,7 @@ class TestImportExportManagerClass(unittest.TestCase):
         numpy.save(file_path_npy, numpy.zeros((4, 4, 4)))
         handler = ImportExportManager.NumPyImportExportHandler("numpy-io-handler", "npy", ["npy"])
         try:
-            data_items = handler.read_data_items(None, "npy", file_path_npy)
+            data_items = handler.read_data_items("npy", pathlib.Path(file_path_npy))
             self.assertEqual(len(data_items), 1)
             data_item = data_items[0]
             self.assertTrue(data_item.large_format)
@@ -351,6 +374,13 @@ class TestImportExportManagerClass(unittest.TestCase):
         finally:
             os.remove(file_path_npy)
 
+    def test_data_item_with_numpy_bool_to_data_element_produces_json_compatible_dict(self):
+        data_item = DataItem.DataItem(numpy.zeros((16, 16)))
+        data_item.large_format = numpy.product((2,3,4), dtype=numpy.int64) > 10  # produces a numpy.bool_
+        with contextlib.closing(data_item):
+            data_element = ImportExportManager.create_data_element_from_data_item(data_item, include_data=False)
+            json.dumps(data_element)
+
 
 if __name__ == '__main__':
     logging.getLogger().setLevel(logging.DEBUG)
